"""Formula 8.2 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MPA
from blueprints.validations import raise_if_negative


class Form8Dot2UltimateBondStress(Formula):
    """Class representing formula 8.2 for the calculation of the design value of the ultimate bond stress for ribbed bars."""

    label = "8.2"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        eta_1: DIMENSIONLESS,
        eta_2: DIMENSIONLESS,
        f_ctd: MPA,
    ) -> None:
        r"""[:math:`f_{bd}`] The design value of the ultimate bond stress for ribbed bars [-].

        NEN-EN 1992-1-1+C2:2011 art.8.4.2(2) - Formula (8.2)

        Parameters
        ----------
        eta_1 : DIMENSIONLESS
            [:math:`η_1`] coefficient related to the quality of the bond condition and the position of the bar during concreting (see Figure 8.2) [-].
            = :math:`1` when ‘good’ conditions are obtained;
            = :math:`1` other cases and for bars in structural elements built with slip-forms, unless it can be shown that ‘good’ bond conditions
            exist;
            Use your own implementation of this formula or use the SubForm8Dot2CoefficientQualityOfBond class.
        eta_2 : DIMENSIONLESS
            [:math:`η_2`] A factor related to the bar diameter [-].
            = :math:`1` for bars with a diameter ≤ :math:`32 \text{mm}`;
            = :math:`(132 - Ø) / 100` for bars with a diameter > :math:`32 \text{mm}`.
            Use your own implementation of this value or use the SubForm8Dot2CoefficientBarDiameter class.
        f_ctd : MPA
            [:math:`f_{ctd}`] Design tensile strength of concrete according to art.3.1.6(2) [MPa].
            Due to the increasing brittleness of higher strength concrete, :math:`f_{ctk,0,05}` should be limited here to the value for C60/75, unless
            it can be verified that the average bond strength increases above this limit.
        """
        super().__init__()
        self.eta_1 = eta_1
        self.eta_2 = eta_2
        self.f_ctd = f_ctd

    @staticmethod
    def _evaluate(
        eta_1: DIMENSIONLESS,
        eta_2: DIMENSIONLESS,
        f_ctd: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd)
        return 2.25 * eta_1 * eta_2 * f_ctd

    def latex(self) -> LatexFormula:
        """Returns a representation of the formula in LaTeX format."""
        return LatexFormula(
            return_symbol=r"f_{bd}",
            result=f"{self:.2f}",
            equation=r"2.25 \cdot \eta_1 \cdot \eta_2 \cdot f_{ctd}",
            numeric_equation=rf"2.25 \cdot {self.eta_1:.2f} \cdot {self.eta_2:.2f} \cdot {self.f_ctd:.2f}",
            comparison_operator_label="=",
        )


class SubForm8Dot2CoefficientQualityOfBond(Formula):
    """Class representing sub-formula for formula 8.2, which calculates the coefficient 'η1' which is dependent on the quality of the bond."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "8.2"

    def __init__(self, bond_quality: str) -> None:
        """[η1] Coefficient that depends on the type of cement [-].

        NEN-EN 1992-1-1+C2:2011 art.8.4.2(2) - η1

        Parameters
        ----------
        bond_quality : str
            Quality of the bond.
                = 'good' for a good bond condition.;
                = 'other' for other cases and for bars in structural elements built with slip-forms, unless it can be shown that ‘good’ bond
                conditions exist.;
        """
        super().__init__()
        self.bond_quality = bond_quality

    @staticmethod
    def _evaluate(bond_quality: str) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        match bond_quality.lower():
            case "good":
                return 1
            case "other":
                return 0.7
            case _:
                raise ValueError(f"Invalid bond quality: {bond_quality}. Options: 'good' or 'other'")


class SubForm8Dot2CoefficientBarDiameter(Formula):
    """Class representing sub-formula for formula 8.2, which calculates the coefficient 'η2' which is dependent on the bar diameter."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "8.2"

    def __init__(self, diameter: MM) -> None:
        """[:math:`η_2`] Coefficient that depends on the bar diameter [-].

        NEN-EN 1992-1-1+C2:2011 art.8.4.2(2) - :math:`η_2`

        Parameters
        ----------
        diameter : MM
            [:math:`Ø`] Diameter of the bar [mm].
        """
        super().__init__()
        self.diameter = diameter

    @staticmethod
    def _evaluate(diameter: MM) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(diameter=diameter)
        if diameter <= 32:
            return 1
        return (132 - diameter) / 100

    def latex(self) -> LatexFormula:
        """Returns a LatexFormula object for this formula."""
        numerical_equation = "1.00" if self.diameter <= 32 else f"(132 - {self.diameter}) / 100"

        return LatexFormula(
            return_symbol=r"\eta_2",
            result=f"{self:.2f}",
            equation=r"\begin{matrix} 1.0 & \text{for }Ø ≤ 32 \\ (132 - Ø) / 100 & \text{for }Ø > 32  \end{matrix}",
            numeric_equation=numerical_equation,
            comparison_operator_label="=",
        )



"""Formula 5.25 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form5Dot25AxialForceCorrectionFactor(Formula):
    """Class representing formula 5.25 for the calculation of the axial force correction factor [:math:`k_{2}`]."""

    label = "5.25"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, n: DIMENSIONLESS) -> None:
        r"""[:math:`k_{2}`] Axial force correction factor.

        NEN-EN 1992-1-1+C2:2011 art.5.8.6(3) - Formula (5.25)

        Parameters
        ----------
        n : DIMENSIONLESS
            [:math:`n`] Relative axial force, :math:`N_{ed} / (A_{c} * f_{cd})` [-].
        """
        super().__init__()
        self.n = n

    @staticmethod
    def _evaluate(n: DIMENSIONLESS) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n=n)
        k2 = n * 0.30
        return min(k2, 0.20)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.25."""
        return LatexFormula(
            return_symbol=r"k_{2}",
            result=f"{self:.3f}",
            equation=r"\min(0.20; n \cdot 0.30)",
            numeric_equation=rf"\min(0.20; {self.n:.3f} \cdot 0.30)",
            comparison_operator_label="=",
        )



"""Stirrups module."""

import numpy as np
from shapely import Point, Polygon
from shapely.geometry.polygon import orient

from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial
from blueprints.type_alias import DIMENSIONLESS, KG_M3, MM, MM2, MM2_M, RATIO
from blueprints.unit_conversion import M_TO_MM, MM3_TO_M3

STIRRUP_COLOR = (0.412, 0.412, 0.412)


class StirrupConfiguration:
    """Representation of a stirrup configuration.

    Parameters
    ----------
    geometry: Polygon
        Line that represents the center-line of the stirrup configuration (clockwise or counterclockwise).
    diameter: MM
        Diameter of the rebar making the stirrup [mm].
    distance: MM
        Longitudinal distance between stirrups [mm].
    material: ReinforcementSteelMaterial
        Reinforcement material.
    shear_check: bool
        Take stirrup into account in shear check
    torsion_check: bool
        Take stirrup into account in torsion check
    mandrel_diameter_factor: MM
        Inner diameter of mandrel as multiple of stirrup diameter [-]
        (default: 4⌀ for ⌀<=16mm and 5⌀ for ⌀>16mm) Tabel 8.1Na NEN-EN 1992-1-1 Dutch National Annex.
    anchorage_length: MM
        Anchorage length [mm]
    based_on_cover: bool
        Default is False. This helps to categorise stirrups that a created based on the covers present in the cross-section.
    relative_start_position: RATIO
        Relative position of the start of the stirrup configuration inside the cross-section (longitudinal direction). Value between 0 and 1.
        Default is 0 (start).
    relative_end_position: RATIO
        Relative position of the end of the stirrup configuration inside the cross-section (longitudinal direction). Value between 0 and 1. Default
        is 1 (end).
    """

    counter = 1

    def __init__(  # noqa: PLR0913
        self,
        geometry: Polygon,
        diameter: MM,
        distance: MM,
        material: ReinforcementSteelMaterial,
        shear_check: bool = True,
        torsion_check: bool = True,
        mandrel_diameter_factor: DIMENSIONLESS | None = None,
        anchorage_length: MM = 0.0,
        based_on_cover: bool = False,
        relative_start_position: RATIO = 0.0,
        relative_end_position: RATIO = 1.0,
        n_vertices_used: int = 4,
        cover_used: MM | None = None,
    ) -> None:
        """Initialisation of the stirrup."""
        self.geometry = orient(polygon=geometry)
        self.diameter = diameter
        self.distance = distance
        self.material = material
        self.shear_check = shear_check
        self.torsion_check = torsion_check
        self.anchorage_length = anchorage_length
        self._mandrel_diameter_factor = mandrel_diameter_factor
        self.based_on_cover = based_on_cover
        self._id = StirrupConfiguration.counter
        self._validation_relative_position(relative_position=relative_start_position)
        self._validation_relative_position(relative_position=relative_end_position)
        self._relative_start_position = relative_start_position
        self._relative_end_position = relative_end_position
        self.n_vertices_used = n_vertices_used
        self._cover_used = cover_used
        self._amount_of_legs = 2
        StirrupConfiguration.counter += 1

    @property
    def mandrel_diameter_factor(self) -> DIMENSIONLESS:
        """Diameter factor of mandrel.
        Standard values given by Dutch Annex Table 8.1Na - NEN-EN 1992-1-1+C2:2011/NB+A1:2020
        (default: 4⌀ for ⌀<=16mm and 5⌀ for ⌀>16mm).
        """
        if self._mandrel_diameter_factor:
            return self._mandrel_diameter_factor
        return 5.0 if self.diameter > 16.0 else 4.0

    @property
    def as_w(self) -> MM2_M:
        """Total cross-sectional area of the stirrup [mm²/m]."""
        return self._amount_of_legs * self.area * (M_TO_MM / self.distance)

    @property
    def area(self) -> MM2:
        """Area of the stirrup bar [mm²]."""
        return 0.25 * np.pi * self.diameter**2

    @property
    def radius(self) -> MM:
        """Radius of the stirrup bar [mm]."""
        return self.diameter / 2

    @property
    def centroid(self) -> Point:
        """Centroid of the stirrup bar [mm]."""
        return self.geometry.centroid

    @property
    def weight_per_meter(self) -> KG_M3:
        """Total mass of the stirrup per meter length in the longitudinal direction (concrete+reinforcement) [kg/m³]
        (Weight of a single stirrup x amount of stirrups present in one meter length).
        """
        return self.material.density * self.geometry.length * self.area * MM3_TO_M3 * M_TO_MM / self.distance

    @property
    def ctc_distance_legs(self) -> MM:
        """Distance between the legs of the stirrup taken form the center lines of the rebar [mm]."""
        min_x, max_x = self.geometry.bounds[0], self.geometry.bounds[2]
        return max_x - min_x

    @property
    def cover_used(self) -> float:
        """Can be used to store the value of the cover used when adding the stirrup to the cross-section [mm]."""
        return self._cover_used or 0.0

    @property
    def relative_start_position(self) -> RATIO:
        """Relative position of the start of the stirrup configuration inside the cross-section. Value between 0 and 1."""
        return self._relative_start_position

    @property
    def relative_end_position(self) -> RATIO:
        """Relative position of the end of the stirrup configuration inside the cross-section. Value between 0 and 1."""
        return self._relative_end_position

    @staticmethod
    def _validation_relative_position(relative_position: DIMENSIONLESS) -> None:
        """Validation of the relative position of the stirrup."""
        if relative_position < 0.0:
            msg = "Relative position of the stirrup must be greater than or equal to zero"
            raise ValueError(msg)
        if relative_position > 1.0:
            msg = "Relative position of the stirrup must be less than or equal to one"
            raise ValueError(msg)

    def __repr__(self) -> str:
        """Representation of the stirrup."""
        return f"Stirrup (id={self._id})|⌀{self.diameter}/{self.material.name}"

    def __str__(self) -> str:
        """String representation of the stirrup."""
        return f"Stirrups ⌀{self.diameter}-{self.distance:.0f} mm | {self.material.name} | {self.as_w:.2f} mm²/m"



"""Formula 5.1 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot1Imperfections(Formula):
    """Class representing formula 5.1 for the calculation of initial inclination imperfections, Θi."""

    label = "5.1"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        theta_0: DIMENSIONLESS,
        alpha_h: DIMENSIONLESS,
        alpha_m: DIMENSIONLESS,
    ) -> None:
        """[Θi] Initial inclination imperfections, Θi, is a ratio between height and inclination of the member [-].

        NEN-EN 1992-1-1+C2:2011 art.5.2(5) - Formula (5.1)

        Parameters
        ----------
        theta_0 : float
            [Θ0] Basic value [-].
            Note: The value of Θ0 for use in a Country may be found in its National Annex.
            The recommended value is 1/200
        alpha_h : float
            [αh] Reduction factor for length or height [-].
            Use your own implementation of this value or use the SubForm5Dot1ReductionFactorLengthOrHeight class.
        alpha_m : float
            [αm] Reduction factor for number of members [-].
            Use your own implementation of this value or use the SubForm5Dot1ReductionFactorNumberOfMembers class.
        """
        super().__init__()
        self.theta_0 = theta_0
        self.alpha_h = alpha_h
        self.alpha_m = alpha_m

    @staticmethod
    def _evaluate(
        theta_0: DIMENSIONLESS,
        alpha_h: DIMENSIONLESS,
        alpha_m: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)
        return theta_0 * alpha_h * alpha_m

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.1."""
        return LatexFormula(
            return_symbol=r"\theta_i",
            result=f"{self:.4f}",
            equation=r"\theta_0 \cdot \alpha_h \cdot \alpha_m",
            numeric_equation=rf"{self.theta_0:.3f} \cdot {self.alpha_h:.3f} \cdot {self.alpha_m:.3f}",
            comparison_operator_label="=",
        )


class SubForm5Dot1ReductionFactorLengthOrHeight(Formula):
    """Class representing sub-formula 5.1 for the calculation of the reduction factor for length or height, αh."""

    label = "5.1"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        length: M,
    ) -> None:
        """[αh] Reduction factor for length or height [-].

        The calculated value of αh is between 2/3 and 1.0.

        NEN-EN 1992-1-1+C2:2011 art.5.2(5) - Formula (5.1)

        Parameters
        ----------
        length : M
            [length] Length or height, see art.5.2(6) [m].
        """
        super().__init__()
        self.length = length

    @staticmethod
    def _evaluate(
        length: M,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(length=length)
        # the value of alpha_h is between 2/3 and 1.0
        alpha_h = 2 / np.sqrt(length)
        if alpha_h < 2 / 3:
            return 2 / 3
        if alpha_h > 1:
            return 1
        return alpha_h

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.1 subformula 1."""
        return LatexFormula(
            return_symbol=r"\alpha_h",
            result=f"{self:.3f}",
            equation=r"\min( \max(2 / \sqrt{l}; 2/3); 1)",
            numeric_equation=rf"\min( \max(2 / \sqrt{{{self.length:.3f}}}; 2/3); 1)",
            comparison_operator_label="=",
        )


class SubForm5Dot1ReductionFactorNumberOfMembers(Formula):
    """Class representing sub-formula 5.1 for the calculation of the reduction factor for number of members, αm."""

    label = "5.1"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        members: int,
    ) -> None:
        """[αm] Reduction factor for number of members [-].

        NEN-EN 1992-1-1+C2:2011 art.5.2(5) - Formula (5.1)

        Parameters
        ----------
        members : int
            [m] Number of vertical members contributing to the total effect [-].
        """
        super().__init__()
        self.members = members

    @staticmethod
    def _evaluate(
        members: int,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(members=members)
        # the value of alpha_m is between 1.0 and 1.5
        return np.sqrt(0.5 * (1 + 1 / members))

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.1 subformula 2."""
        return LatexFormula(
            return_symbol=r"\alpha_m",
            result=f"{self:.3f}",
            equation=r"\sqrt{0.5 \cdot ( 1 + 1 / m)}",
            numeric_equation=rf"\sqrt{{0.5 \cdot ( 1 + 1 / {self.members:.3f})}}",
            comparison_operator_label="=",
        )



"""Formula 5.35 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot35EffectiveDepth(Formula):
    r"""Class representing formula 5.35 for the calculation of the effective depth, :math:`d`."""

    label = "5.35"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        h: MM,
        i_s: MM,
    ) -> None:
        r"""[:math:`d`] Effective depth [:math:`mm`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.8.3 - Formula (5.35)

        Parameters
        ----------
        h : MM
            [:math:`h`] Total height of the section [:math:`mm`].
        i_s : MM
            [:math:`i_s`] Radius of gyration of the total reinforcement area [:math:`mm`].
        """
        super().__init__()
        self.h = h
        self.i_s = i_s

    @staticmethod
    def _evaluate(
        h: MM,
        i_s: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(h=h, i_s=i_s)

        return h / 2 + i_s

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.35."""
        return LatexFormula(
            return_symbol=r"d",
            result=f"{self:.3f}",
            equation=r"\frac{h}{2} + i_s",
            numeric_equation=rf"\frac{{{self.h:.3f}}}{{2}} + {self.i_s:.3f}",
            comparison_operator_label="=",
            unit="mm",
        )



"""Formula 5.24 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form5Dot24AxialForceCorrectionFactor(Formula):
    """Class representing formula 5.24 for the calculation of the axial force correction factor [:math:`k_{2}`]."""

    label = "5.24"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        n: DIMENSIONLESS,
        lambda_factor: DIMENSIONLESS,
    ) -> None:
        """[:math:`k_{2}`] Axial force correction factor.

        NEN-EN 1992-1-1+C2:2011 art.5.8.3 - Formula (5.24)

        Parameters
        ----------
        n : DIMENSIONLESS
            [:math:`n`] Relative axial force, :math:`N_{ed} / (A_{c} * f_{cd})` [-].
        lambda_factor : DIMENSIONLESS
            [:math:`λ`] Slenderness ratio, see 5.8.3 [-].

        Notes
        -----
        The value of :math:`k_{2}` cannot be larger than 0.20.
        """
        super().__init__()
        self.n = n
        self.lambda_factor = lambda_factor

    @staticmethod
    def _evaluate(n: DIMENSIONLESS, lambda_factor: DIMENSIONLESS) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n=n, lambda_factor=lambda_factor)
        k2 = n * lambda_factor / 170
        return min(k2, 0.20)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.24."""
        return LatexFormula(
            return_symbol=r"k_2",
            result=f"{self:.3f}",
            equation=r"\min(0.20; n \cdot \frac{λ}{170})",
            numeric_equation=rf"\min(0.20; {self.n:.3f} \cdot \frac{{{self.lambda_factor:.3f}}}{{170}})",
            comparison_operator_label="=",
        )



"""Generic module for the concrete structural class according to NEN-EN 1992-1-1 Concrete - General."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Self

from blueprints.codes.eurocode.exposure_classes import ExposureClassesBase as ExposureClasses
from blueprints.materials.concrete import ConcreteMaterial
from blueprints.type_alias import YEARS


class ConcreteStructuralClassBase(int):
    """Base class for the Structural class of the concrete element.

    This Class can be used to keep track of the operations of the structural class of the concrete element.

    In accordance with:
    NEN-EN 1992-1-1 Concrete - General
    """

    def __new__(
        cls,
        concrete_structural_class_calculator: type[AbstractConcreteStructuralClassCalculator],
        *args,
        **kwargs,
    ) -> Self:
        """Constructor of the ConcreteStructuralClassBase class.

        Parameters
        ----------
        concrete_structural_class_calculator: type[AbstractConcreteStructuralClassCalculator]
            The calculator class for the structural class of the concrete element
        *args
            Arguments for initiating the calculator instance
        **kwargs
            Keyword arguments for initiating the calculator instance

        Raises
        ------
        TypeError
            If concrete_structural_class_calculator is not a subclass of AbstractConcreteStructuralClassCalculator

        Returns
        -------
        StructuralClass
            The instance of the ConcreteStructuralClassBase class
        """
        if not issubclass(concrete_structural_class_calculator, AbstractConcreteStructuralClassCalculator):
            raise TypeError(f"unsupported calculator type(s) for the ConcreteStructuralClassBase class: '{concrete_structural_class_calculator}'")
        calculator_instance = concrete_structural_class_calculator(*args, **kwargs)
        calculator_instance.calculate_structural_class()
        structural_class = calculator_instance.structural_class
        new_instance = super().__new__(cls, structural_class)
        new_instance.explanation = calculator_instance.explanation
        return new_instance

    @property
    def explanation(self) -> str:
        """Property which returns the structural class explanation.

        Returns
        -------
        str
            The explanation of the structural class
        """
        return self._explanation

    @explanation.setter
    def explanation(self, explanation: str) -> None:
        """Setter for the explanation of the structural class.

        Parameters
        ----------
        explanation: str
            The explanation of the structural class

        Raises
        ------
        AttributeError
            If the explanation is already set
        """
        if not hasattr(self, "_explanation"):
            self._explanation = explanation
        else:
            raise AttributeError(f"Attribute 'explanation' of {self.__class__} object is read-only and cannot be modified after initialization.")

    def __str__(self) -> str:
        """Returns a string representation of the object.

        Overrides the default __str__ method to include the prefix 'S' before the string representation
        of the superclass.

        Returns
        -------
        str
            The string representation of the object.
        """
        return f"S{super().__str__()}"


class AbstractConcreteStructuralClassCalculator(ABC):
    """(Abstract class for) Structural class calculator of the concrete element.

    This abstract class is in accordance with:
    NEN-EN 1992-1-1 Concrete - General

    This abstract class should be implemented in for each specific release of the Eurocode.
    """

    DEFAULT_STRUCTURAL_CLASS = 0
    DEFAULT_EXPLANATION = ""

    def __init__(
        self,
        exposure_classes: ExposureClasses,
        design_working_life: YEARS,
        concrete_material: ConcreteMaterial,
        plate_geometry: bool,
        quality_control: bool,
    ) -> None:
        """Initializer of the AbstractConcreteStructuralClassCalculator class.

        The concrete implementation of this class will be used to calculate the structural class.

        Parameters
        ----------
        exposure_classes : ExposureClasses
            The exposure classes of the concrete element
        design_working_life : YEARS
            The design working life of the concrete element
        concrete_material : ConcreteMaterial
            The concrete material of the concrete element
        plate_geometry : bool
            True if the concrete element has a plate geometry, False otherwise
        quality_control : bool
            True if the quality control of the concrete element is ensured, False otherwise
        """
        self.exposure_classes = exposure_classes
        self.design_working_life = design_working_life
        self.concrete_material = concrete_material
        self.plate_geometry = plate_geometry
        self.quality_control = quality_control
        self._calculated = False
        self._structural_class = self.DEFAULT_STRUCTURAL_CLASS
        self._explanation = self.DEFAULT_EXPLANATION

    def __setattr__(self, name: str, value: Any) -> None:  # noqa: ANN401
        """Setter for the attributes of the class.

        This method is used to set the attributes of the class and check if the structural class needs to be recalculated.

        Parameters
        ----------
        name: str
            The name of the attribute
        value: Any
            The value of the attribute
        """
        attributes_affecting_calculation = {"exposure_classes", "design_working_life", "concrete_material", "plate_geometry", "quality_control"}
        if name in attributes_affecting_calculation and getattr(self, name, None) != value:
            self._calculated = False
        super().__setattr__(name, value)

    @property
    @abstractmethod
    def source_document(self) -> str:
        """Property for the source document.

        For example, "NEN-EN 1992-1-1+C2:2011"
        Try to use the official and complete name of the document including publishing year, if possible.

        Returns
        -------
        str
            The reference to the document where the structural class calculation method originates.
            This is an abstract method and must be implemented in all subclasses.
        """

    @property
    def structural_class(self) -> int:
        """Property which returns the structural class.

        Returns
        -------
        int
            the structural class
        """
        if not self._calculated:
            raise ValueError("The structural class has not been calculated yet.")
        return self._structural_class

    @property
    def explanation(self) -> str:
        """Property which returns the structural class explanation."""
        if not self._calculated:
            raise ValueError("The structural class has not been calculated yet.")
        return self._explanation

    def update_structural_class(self, delta: int, explanation: str) -> None:
        """Method to update the structural class with the given delta and explanation.

        Parameters
        ----------
        delta: int
            The delta to be added to the structural class
        explanation: str
            The explanation of the structural class calculation

        Raises
        ------
        TypeError
            If the delta is not an integer
        """
        if not isinstance(delta, int):
            raise TypeError(f"unsupported delta type(s) for the update operation: '{type(delta)}'")
        unit_suffix = "classes" if abs(delta) > 1 else "class"
        operator_symbol = "+" if delta >= 0 else "-"
        self._structural_class += delta
        self._explanation = f"{self._explanation} {operator_symbol} {abs(delta)} {unit_suffix} ({explanation})"

    def calculate_structural_class(
        self,
    ) -> None:
        """Method to execute the calculation of the structural class."""
        if not self._calculated:
            self._structural_class_delta_design_working_life()
            self._structural_class_delta_concrete_grade()
            self._structural_class_delta_plate_geometry()
            self._structural_class_delta_quality_control()
            self._calculated = True

    @abstractmethod
    def _structural_class_delta_design_working_life(self) -> None:
        """Calculates the addition to the structural class based on the design working life."""

    @abstractmethod
    def _structural_class_delta_quality_control(self) -> None:
        """Calculates the addition to the structural class based on the quality control."""

    @abstractmethod
    def _structural_class_delta_plate_geometry(self) -> None:
        """Calculates the addition to the structural class based on the plate geometry."""

    @abstractmethod
    def _structural_class_delta_concrete_grade(self) -> None:
        """Calculates the addition to the structural class based on the concrete grade."""



"""Base class for constants for the calculation of nominal concrete cover."""

from abc import ABC, abstractmethod
from dataclasses import dataclass

from blueprints.checks.nominal_concrete_cover.definitions import AbrasionClass, CastingSurface
from blueprints.type_alias import MM


@dataclass(frozen=True)
class NominalConcreteCoverConstantsBase(ABC):
    """Base class for constants for the calculation of nominal concrete cover."""

    CODE_SUFFIX: str  # Suffix for the code, e.g. "+C2:2011"
    COVER_INCREASE_FOR_UNEVEN_SURFACE: MM
    COVER_INCREASE_FOR_ABRASION_CLASS: dict[AbrasionClass, MM]
    DEFAULT_DELTA_C_DEV: MM

    @staticmethod
    @abstractmethod
    def minimum_cover_with_regard_to_casting_surface(c_min_dur: MM, casting_surface: CastingSurface) -> MM:
        """Calculate the minimum cover with regard to casting surface according to art. 4.4.1.3 (4) from NEN-EN 1992-1-1."""

    @staticmethod
    @abstractmethod
    def minimum_cover_with_regard_to_casting_surface_latex(casting_surface: CastingSurface) -> str:
        """LateX representation of minimum cover with regard to casting surface according to art. 4.4.1.3 (4) from NEN-EN 1992-1-1."""



"""Latex formula representation."""

from dataclasses import dataclass


@dataclass(frozen=True)
class LatexFormula:
    """Latex formula representation.

    Depending on the context this could include the unit, the formula, the result, etc.

    Attributes
    ----------
    return_symbol: str
        The symbol to return
    result: str
        The result of the formula
    equation: str, default ""
        The formula with symbols
    numeric_equation: str, default ""
        The formula with values (numbers)
    comparison_operator_label: str, default "="
        The label for the comparison operators between the return symbol and the result.
        Could be changed for inequalities.
    """

    return_symbol: str
    result: str
    equation: str = ""
    numeric_equation: str = ""
    comparison_operator_label: str = "="
    unit: str = ""

    @property
    def complete(self) -> str:
        """Complete representation of the formula.

        Returns
        -------
        str
            Return symbol = equation = numeric_equation = result

        """
        all_sub_equations = [self.return_symbol, self.equation, self.numeric_equation, f"{self.result}"]
        long_formula = f" {self.comparison_operator_label} ".join([eq for eq in all_sub_equations if eq != ""])
        return long_formula + f" {self.unit}" if self.unit else long_formula

    @property
    def short(self) -> str:
        """Minimal representation of the formula.

        Returns
        -------
        str
            Return symbol = result

        """
        short_formula = f"{self.return_symbol} {self.comparison_operator_label} {self.result}"
        return short_formula + f" {self.unit}" if self.unit else short_formula

    def __str__(self) -> str:
        """String representation of the formula."""
        return self.complete


def latex_fraction(numerator: str | float, denominator: str | float) -> str:
    r"""Return a string which will output: \frac{numerator}{denominator} in latex.

    Examples
    --------
    >>> latex_fraction(1, 2)
    str(\frac{1}{2})

    Parameters
    ----------
    numerator: str | float
        The numerator of the fraction.
    denominator: str | float
        The denominator of the fraction.

    Returns
    -------
    str
        The latex string

    """
    return f"\\frac{{{numerator}}}{{{denominator}}}"


def latex_min_curly_brackets(*args: str | float) -> str:
    r"""Return a string which will output: min{arg_1; arg_2; ...; arg_N} in latex.

    It will also automatically ensure floats are converted to latex
    text.

    Examples
    --------
    >>> latex_min_curly_brackets(1, 2)
    str(\min \left\{1; 2\right\})

    Parameters
    ----------
    args: str
        The arguments of the min function.

    Returns
    -------
    str
        The latex representation of the min function.
    """
    arguments = [str(arg) for arg in args]
    return f"\\min \\left\\{{{'; '.join(arguments)}\\right\\}}"


def latex_max_curly_brackets(*args: str | float) -> str:
    r"""Return a string which will output: max{arg_1; arg_2; ...; arg_N} in latex.

    It will also automatically ensure floats are converted to latex
    text.

    Examples
    --------
    >>> latex_max_curly_brackets(1, 2)
    str(\max \left\{1; 2\right\})

    Parameters
    ----------
    args: str
        The arguments of the max function.

    Returns
    -------
    str
        The latex representation of the max function.
    """
    arguments = [str(arg) for arg in args]
    return f"\\max \\left\\{{{'; '.join(arguments)}\\right\\}}"


def latex_replace_symbols(template: str, replacements: dict[str, str], unique_symbol_check: bool = True) -> str:
    r"""
    Replace symbols in a LaTeX template string based on the provided dictionary.

    This function searches the template for symbols specified in the
    replacements and replaces them with their corresponding
    values. It also checks for the occurrence of symbols based on the
    unique_symbol_check parameter, raising an error if necessary.

    Examples
    --------
    >>> latex_template = r"\frac{K_{MOD}}{B}"
    >>> replacements = {"K_{MOD}": "1.0", "B": "y"}
    >>> latex_replace_symbols(latex_template, replacements)
    '\frac{1.0}{y}

    Parameters
    ----------
    template: str
        The original LaTeX string containing symbols to replace.

    replacements: dict[str, str]
        A dictionary where keys are symbols to be replaced and values
        are their replacements.

    unique_symbol_check: bool, optional
        If True (default), raises an error if a symbol appears more
        than once in the template. If False, multiple occurrences
        will be replaced without error.

    Returns
    -------
    str
        The modified LaTeX string with symbols replaced.

    Raises
    ------
    ValueError
        If a symbol in the dictionary is not found in the template,
        or if a symbol appears more than once when single_symbol_search is True.
    """
    _filled_latex_string: str = template
    for symbol, replacement in replacements.items():
        occurrences = _filled_latex_string.count(symbol)

        # Check for the presence of the symbol in the template
        if occurrences == 0:
            raise ValueError(f"Symbol '{symbol}' not found in the template.")

        # If single_symbol_search is True, check for multiple occurrences
        if unique_symbol_check and occurrences > 1:
            raise ValueError(f"Symbol '{symbol}' found multiple times in the template.")

        # Replace the symbol with its replacement
        _filled_latex_string = _filled_latex_string.replace(symbol, replacement)

    return _filled_latex_string



"""Formula 5.36 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN, MM2, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot36RelativeAxialForce(Formula):
    r"""Class representing formula 5.36 for the calculation of the relative axial force, :math:`K_{r}`."""

    label = "5.36"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        n_ed: KN,
        ac: MM2,
        fcd: MPA,
        as_: MM2,
        fyd: MPA,
        n_bal: DIMENSIONLESS,
    ) -> None:
        r"""[:math:`K_{r}`] Relative axial force [:math:`DIMENSIONLESS`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.8.3(3) - Formula (5.36)

        Parameters
        ----------
        n_ed : KN
            [:math:`N_{Ed}`] Design value of axial load [:math:`kN`].
        ac : float
            [:math:`A_{c}`] Area of concrete cross section [:math:`mm^2`].
        fcd : float
            [:math:`f_{cd}`] Design value of concrete compressive strength [:math:`MPa`].
        as_ : float
            [:math:`A_{s}`] Total area of reinforcement [:math:`mm^2`].
        fyd : float
            [:math:`f_{yd}`] Design yield strength of reinforcement [:math:`MPa`].
        n_bal : DIMENSIONLESS
            [:math:`n_{bal}`] Value of n at maximum moment resistance, 0.4 may be used [-].
        """
        super().__init__()
        self.n_ed = n_ed
        self.ac = ac
        self.fcd = fcd
        self.as_ = as_
        self.fyd = fyd
        self.n_bal = n_bal

    @staticmethod
    def _evaluate(
        n_ed: KN,
        ac: MM2,
        fcd: MPA,
        as_: MM2,
        fyd: MPA,
        n_bal: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n_ed=n_ed, as_=as_, fyd=fyd, n_bal=n_bal)
        raise_if_less_or_equal_to_zero(ac=ac, fcd=fcd)

        omega = as_ * fyd / (ac * fcd)
        nu = 1 + omega
        n = n_ed / (ac * fcd)
        return min((nu - n) / (nu - n_bal), 1)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.36."""
        return LatexFormula(
            return_symbol=r"K_{r}",
            result=f"{self:.3f}",
            equation=r"\min\left(\frac{\left(1 + \frac{A_{s} \cdot f_{yd}}{A_{c} \cdot f_{cd}}\right) - "
            r"\frac{N_{Ed}}{A_{c} \cdot f_{cd}}}{\left(1 + \frac{A_{s} \cdot f_{yd}}{A_{c} \cdot f_{cd}}\right) - n_{bal}}, 1\right)",
            numeric_equation=rf"\min\left(\frac{{\left(1 + \frac{{{self.as_:.3f} \cdot {self.fyd:.3f}}}{{{self.ac:.3f} \cdot "
            rf"{self.fcd:.3f}}}\right) - \frac{{{self.n_ed:.3f}}}{{{self.ac:.3f} \cdot {self.fcd:.3f}}}}}{{\left(1 + \frac{{{self.as_:.3f} \cdot "
            rf"{self.fyd:.3f}}}{{{self.ac:.3f} \cdot {self.fcd:.3f}}}\right) - {self.n_bal:.3f}}}, 1\right)",
            comparison_operator_label="=",
            unit="-",
        )



"""Module for the concrete exposure classes
according to Table 4.1 from NEN-EN 1992-1-1+C2:2011: Chapter 4 - Durability and cover to reinforcement.
"""

from collections.abc import Sequence
from functools import total_ordering
from typing import Self, TypeVar

from blueprints.codes.eurocode.exposure_classes import (
    CarbonationBase,
    ChemicalBase,
    ChlorideBase,
    ChlorideSeawaterBase,
    ExposureClassesBase,
    FreezeThawBase,
)


@total_ordering
class Carbonation(CarbonationBase):
    """Enum Class which indicates the classification of corrosion induced by carbonation."""

    NA = "Not applicable"
    XC1 = "XC1"
    XC2 = "XC2"
    XC3 = "XC3"
    XC4 = "XC4"

    @staticmethod
    def exposure_class_description() -> str:
        """Static method which returns the description of this exposure class.

        Returns
        -------
        str
            description of this exposure class
        """
        return "Corrosion induced by carbonation"

    def description_of_the_environment(self) -> str:
        """Description of the environment based on the instance.

        Returns
        -------
        str
            description of the environment based on the instance
        """
        match self:
            case Carbonation.XC1:
                return "Dry or permanently wet"
            case Carbonation.XC2:
                return "Wet, rarely dry"
            case Carbonation.XC3:
                return "Moderate humidity"
            case Carbonation.XC4:
                return "Cyclic wet and dry"
            case Carbonation.NA:
                return "Not applicable"

    @staticmethod
    def notation() -> str:
        """Static method which returns the notation of this exposure class.

        Returns
        -------
        str
            notation of this exposure class
        """
        return "XC"


@total_ordering
class Chloride(ChlorideBase):
    """Enum Class which indicates the classification of corrosion induced by chlorides other than by sea water."""

    NA = "Not applicable"
    XD1 = "XD1"
    XD2 = "XD2"
    XD3 = "XD3"

    @staticmethod
    def exposure_class_description() -> str:
        """Static method which returns the description of this exposure class.

        Returns
        -------
        str
            description of this exposure class
        """
        return "Corrosion induced by chlorides other than by sea water"

    def description_of_the_environment(self) -> str:
        """Description of the environment based on the instance.

        Returns
        -------
        str
            description of the environment based on the instance
        """
        match self:
            case Chloride.XD1:
                return "Moderate humidity"
            case Chloride.XD2:
                return "Wet, rarely dry"
            case Chloride.XD3:
                return "Cyclic wet and dry"
            case Chloride.NA:
                return "Not applicable"

    @staticmethod
    def notation() -> str:
        """Static method which returns the notation of this exposure class.

        Returns
        -------
        str
            notation of this exposure class
        """
        return "XD"


@total_ordering
class ChlorideSeawater(ChlorideSeawaterBase):
    """Enum Class which indicates the classification of corrosion induced by chlorides from sea water."""

    NA = "Not applicable"
    XS1 = "XS1"
    XS2 = "XS2"
    XS3 = "XS3"

    @staticmethod
    def exposure_class_description() -> str:
        """Static method which returns the description of this exposure class.

        Returns
        -------
        str
            description of this exposure class
        """
        return "Corrosion induced by chlorides from sea water"

    def description_of_the_environment(self) -> str:
        """Description of the environment based on the instance.

        Returns
        -------
        str
            description of the environment based on the instance
        """
        match self:
            case ChlorideSeawater.XS1:
                return "Exposed to airborne salt but not in direct contact with sea water"
            case ChlorideSeawater.XS2:
                return "Permanently submerged"
            case ChlorideSeawater.XS3:
                return "Tidal, splash and spray zones"
            case ChlorideSeawater.NA:
                return "Not applicable"

    @staticmethod
    def notation() -> str:
        """Static method which returns the notation of this exposure class.

        Returns
        -------
        str
            notation of this exposure class
        """
        return "XS"


@total_ordering
class FreezeThaw(FreezeThawBase):
    """Enum Class which indicates the classification of freeze/thaw attack with or without de-icing agents."""

    NA = "Not applicable"
    XF1 = "XF1"
    XF2 = "XF2"
    XF3 = "XF3"
    XF4 = "XF4"

    @staticmethod
    def exposure_class_description() -> str:
        """Static method which returns the description of this exposure class.

        Returns
        -------
        str
            description of this exposure class
        """
        return "Freeze/thaw attack with or without de-icing agents"

    def description_of_the_environment(self) -> str:
        """Description of the environment based on the instance.

        Returns
        -------
        str
            description of the environment based on the instance
        """
        match self:
            case FreezeThaw.XF1:
                return "Moderate water saturation, without de-icing agent"
            case FreezeThaw.XF2:
                return "Moderate water saturation, with de-icing agent"
            case FreezeThaw.XF3:
                return "High water saturation, without de-icing agents"
            case FreezeThaw.XF4:
                return "High water saturation with de-icing agents or sea water"
            case FreezeThaw.NA:
                return "Not applicable"

    @staticmethod
    def notation() -> str:
        """Static method which returns the notation of this exposure class.

        Returns
        -------
        str
            notation of this exposure class
        """
        return "XF"


@total_ordering
class Chemical(ChemicalBase):
    """Enum Class which indicates the classification of chemical attack."""

    NA = "Not applicable"
    XA1 = "XA1"
    XA2 = "XA2"
    XA3 = "XA3"

    @staticmethod
    def exposure_class_description() -> str:
        """Static method which returns the description of this exposure class.

        Returns
        -------
        str
            description of this exposure class
        """
        return "Chemical attack"

    def description_of_the_environment(self) -> str:
        """Description of the environment based on the instance.

        Returns
        -------
        str
            description of the environment based on the instance
        """
        match self:
            case Chemical.XA1:
                return "Slightly aggressive chemical environment according to EN 206-1, Table 2"
            case Chemical.XA2:
                return "Moderately aggressive chemical environment according to EN 206-1, Table 2"
            case Chemical.XA3:
                return "Highly aggressive chemical environment according to EN 206-1, Table 2"
            case Chemical.NA:
                return "Not applicable"

    @staticmethod
    def notation() -> str:
        """Static method which returns the notation of this exposure class.

        Returns
        -------
        str
            notation of this exposure class
        """
        return "XA"


# Define a generic type for the class
T = TypeVar("T", bound="Table4Dot1ExposureClasses")


class Table4Dot1ExposureClasses(ExposureClassesBase):
    """Implementation of table 4.1 from NEN-EN 1992-1-1+C2:2011.

    Exposure classes related to environmental conditions in accordance with EN 206-1
    """

    carbonation: Carbonation
    chloride: Chloride
    chloride_seawater: ChlorideSeawater
    freeze: FreezeThaw
    chemical: Chemical

    @classmethod
    def from_exposure_list(cls, exposure_classes: Sequence[str]) -> Self:
        """Create an instance from a sequence of exposure classes.

        Examples
        --------
        >>> exposure_classes = ["XC1", "XD1", "XS1"]
        >>> Table4Dot1ExposureClasses().from_exposure_list(exposure_classes)
        Table4Dot1ExposureClasses(
            carbonation=<Carbonation.XC1: 'XC1'>,
            chloride=<Chloride.XD1: 'XD1'>,
            chloride_seawater=<ChlorideSeawater.XS1: 'XS1'>,
            freeze=<FreezeThaw.NA: 'Not applicable'>,
            chemical=<Chemical.NA: 'Not applicable'>
        )

        Parameters
        ----------
        exposure_classes : Sequence[str]
            list of exposure classes, order is not important. If an exposure class is not provided, it is set to "Not applicable"
            You can use capital letters or lowercase letters, the method is case-insensitive.
            For example, "XC1" and "xc1" are both valid.

        Returns
        -------
        Self
            instance created from the list
        """
        exposures: dict[str, Carbonation | Chloride | ChlorideSeawater | Chemical | FreezeThaw] = {}
        classifications = {
            classification.notation(): classification for classification in (Carbonation, Chloride, ChlorideSeawater, FreezeThaw, Chemical)
        }

        for exposure_str in exposure_classes:
            classification = classifications.get(exposure_str[:2].upper())
            if classification is None:
                raise ValueError(f"Invalid exposure class: '{exposure_str}'")
            classification_name = classification.snake_case().removesuffix("_thaw")
            if classification_name in exposures:
                raise ValueError(f"Duplication Error: There are multiple instances of '{classification.__name__}' class.")
            exposures[classification_name] = classification[exposure_str.upper()]

        for classification in classifications.values():
            classification_name = classification.snake_case().removesuffix("_thaw")
            if classification_name not in exposures:
                exposures.setdefault(classification_name, classification.NA)

        return cls(**exposures)  # type: ignore[arg-type]



"""Formula 3.3 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA


class Form3Dot3AxialTensileStrengthFromTensileSplittingStrength(Formula):
    """Class representing formula 3.3 for the approximated axial tensile strength, fct, determined by tensile splitting strength."""

    label = "3.3"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ct_sp: MPA,
    ) -> None:
        """[fct] The approximated axial tensile strength when tensile strength is determined as coefficient
        which is dependent of the age of concrete [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.2(8) - Formula (3.3)

        Parameters
        ----------
        f_ct_sp : float
            [fct,sp] Tensile strength determined by tensile splitting strength [MPa].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ct_sp = f_ct_sp

    @staticmethod
    def _evaluate(
        f_ct_sp: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_ct_sp < 0:
            raise ValueError(f"Negative f_ct_sp: {f_ct_sp}. f_ct_sp cannot be negative")
        return 0.9 * f_ct_sp

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.3."""
        return LatexFormula(
            return_symbol=r"f_{ct}",
            result=f"{self:.3f}",
            equation=r"0.9 \cdot f_{ct,sp}",
            numeric_equation=rf"0.9 \cdot {self.f_ct_sp:.3f}",
            comparison_operator_label="=",
        )



"""Formula 5.34 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot34Curvature(Formula):
    r"""Class representing formula 5.34 for the calculation of the curvature, :math:`\frac{1}{r}`."""

    label = "5.34"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        k_r: DIMENSIONLESS,
        k_phi: DIMENSIONLESS,
        f_yd: MPA,
        e_s: MPA,
        d: MM,
    ) -> None:
        r"""[:math:`\frac{1}{r}`] Curvature [:math:`1/mm`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.8.3 - Formula (5.34)

        Parameters
        ----------
        k_r : DIMENSIONLESS
            [:math:`K_r`] Correction factor depending on axial load, see 5.8.8.3 (3) [-].
        k_phi : DIMENSIONLESS
            [:math:`K_\\phi`] Factor for taking account of creep, see 5.8.8.3 (4) [-].
        f_yd : MPA
            [:math:`f_{yd}`] Design yield strength of reinforcement [:math:`MPa`].
        e_s : MPA
            [:math:`E_s`] Modulus of elasticity of reinforcement [:math:`MPa`].
        d : MM
            [:math:`d`] Effective depth; see also 5.8.8.3 (2) [:math:`mm`].
        """
        super().__init__()
        self.k_r = k_r
        self.k_phi = k_phi
        self.f_yd = f_yd
        self.e_s = e_s
        self.d = d

    @staticmethod
    def _evaluate(
        k_r: DIMENSIONLESS,
        k_phi: DIMENSIONLESS,
        f_yd: MPA,
        e_s: MPA,
        d: MM,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            k_r=k_r,
            k_phi=k_phi,
        )
        raise_if_less_or_equal_to_zero(f_yd=f_yd, e_s=e_s, d=d)

        return k_r * k_phi * (f_yd / e_s) / (0.45 * d)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.34."""
        return LatexFormula(
            return_symbol=r"\frac{1}{r}",
            result=f"{self:.6f}",
            equation=r"K_r \cdot K_\phi \cdot \frac{f_{yd}}{E_s \cdot 0.45 \cdot d}",
            numeric_equation=rf"{self.k_r:.3f} \cdot {self.k_phi:.3f} \cdot \frac{{{self.f_yd:.3f}}}{{{self.e_s:.3f} \cdot 0.45 \cdot {self.d:.3f}}}",
            comparison_operator_label="=",
            unit="1/mm",
        )



"""Formula 8.20 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form8Dot20BondStrengthAnchorageULS(Formula):
    """Class representing formula 8.20 for the calculation of bond strength for anchorage in the ultimate limit state :math:`f_{bpd}`.

    NEN-EN 1992-1-1+C2:2011 art.8.10.2.2(3) - Formula (8.20)

    Parameters
    ----------
    eta_p2 : DIMENSIONLESS
        [:math:`η_{p2}`] Coefficient that takes into account the type of tendon and the bond situation at anchorage [:math:`-`].

        1.4 for indented wires or 1,2 for 7-wire strands.
    eta_1 : DIMENSIONLESS
        [:math:`η_{1}`] Coefficient for concrete, defined in 8.10.2.2 (1) [:math:`-`].
    f_ctd : MPA
        Design tensile strength of concrete [:math:`MPa`].
    """

    label = "8.20"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, eta_p2: DIMENSIONLESS, eta_1: DIMENSIONLESS, f_ctd: MPA) -> None:
        super().__init__()
        self.eta_p2 = eta_p2
        self.eta_1 = eta_1
        self.f_ctd = f_ctd

    @staticmethod
    def _evaluate(eta_p2: DIMENSIONLESS, eta_1: DIMENSIONLESS, f_ctd: MPA) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(eta_p2=eta_p2, eta_1=eta_1, f_ctd=f_ctd)
        return eta_p2 * eta_1 * f_ctd

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 8.20."""
        return LatexFormula(
            return_symbol=r"f_{bpd}",
            result=f"{self:.3f}",
            equation=r"\eta_{p2} \cdot \eta_{1} \cdot f_{ctd}",
            numeric_equation=rf"{self.eta_p2:.3f} \cdot {self.eta_1:.3f} \cdot {self.f_ctd:.3f}",
            comparison_operator_label="=",
        )



"""Formula 9.12N from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailling and specific rules."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, MM2, MPA
from blueprints.unit_conversion import KN_TO_N
from blueprints.validations import raise_if_negative


class Form9Dot12nMinimumLongitudinalReinforcementColumns(Formula):
    """Class representing the formula 9.12N for the calculation of the minimum longitudinal reinforcement for columns."""

    label = "9.12N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        n_ed: KN,
        f_yd: MPA,
        a_c: MM2,
    ) -> None:
        """[As,min] Minimum longitudinal reinforcement for columns [mm²].

        NEN-EN 1992-1-1+C2:2011 art.9.5.2(2) - Formula (9.12N)

        Parameters
        ----------
        n_ed: KN
            [Ned] Design value of axial force [kN].
        f_yd: MPA
            [fyd] Design yield strength reinforcement steel [MPa].
        a_c: MM2
            [Ac] Concrete cross-sectional area [mm²].
        """
        super().__init__()
        self.n_ed = n_ed
        self.f_yd = f_yd
        self.a_c = a_c

    @staticmethod
    def _evaluate(
        n_ed: KN,
        f_yd: MPA,
        a_c: MM2,
    ) -> MM2:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(n_ed=n_ed, f_yd=f_yd, a_c=a_c)
        return max(0.1 * n_ed * KN_TO_N / f_yd, 0.002 * a_c)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 9.12N."""
        return LatexFormula(
            return_symbol=r"A_{s,min}",
            result=f"{self:.2f}",
            equation=r"\max( \frac{0.10 \cdot N_{Ed}}{f_{yd}}, 0.002 \cdot A_c )",
            numeric_equation=rf"\max( \frac{{0.10 \cdot {self.n_ed:.2f}}}{{{self.f_yd:.2f}}}, 0.002 \cdot {self.a_c:.2f} )",
            comparison_operator_label="=",
        )



"""Formula 2.4 from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

import math

from blueprints.codes.eurocode.nen_9997_1_c2_2017 import NEN_9997_1_C2_2017
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN
from blueprints.validations import raise_if_negative


class Form2Dot4DesignValueGeotechnicalParameter:
    """Class representing formula 2.4 for the check of
    the destabilizing load effect
    against the stabilizing load effect and friction resistance
    :math:`E_{dst;d} ≤ E_{stb;d} + T_d`.
    """

    label = "2.4"
    source_document = NEN_9997_1_C2_2017

    def __init__(self, e_dst_d: KN, e_stb_d: KN, t_d: KN) -> None:
        """Check of the destabilizing load effect
        against the stabilizing load effect and friction resistance
        [:math:`E_{dst;d} ≤ E_{stb;d} + T_d`].

        NEN 9997-1+C2:2017 art.2.4.7.2(1) - Formula (2.4)

        Parameters
        ----------
        E_dst_d : N
            [:math:`E_dst;d`] Design value of destabilizing load effect.
        E_stb_d : N
            [:math:`E_stb;d`] Design value of stabilizing load effect.
        T_d : N
            [:math: `T_d`] Design value of friction resistance.
        """
        self.e_dst_d = e_dst_d
        self.e_stb_d = e_stb_d
        self.t_d = t_d

    def __bool__(self) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(e_dst_d=self.e_dst_d, e_stb_d=self.e_stb_d)
        return self.e_dst_d < self.e_stb_d + self.t_d or math.isclose(self.e_dst_d, self.e_stb_d + self.t_d)

    def __str__(self) -> str:
        """Returns a string representation of the formula."""
        return self.latex().complete

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 2.4."""
        return LatexFormula(
            return_symbol="",
            equation="E_{dst;d} \\leq E_{stb;d} + T_d",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            numeric_equation=f"{self.e_dst_d:.2f} \\leq {self.e_stb_d:.2f} + {self.t_d:.2f} \\to {self.e_dst_d:.2f} \\leq "
            f"{self.e_stb_d + self.t_d:.2f}",
            comparison_operator_label="\\to",
        )



"""Formula 5.16 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

import math

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_negative


class Form5Dot16EffectiveLengthUnbraced(Formula):
    """Class representing formula 5.16 for the calculation of effective length of unbraced members, :math:`l_0`."""

    label = "5.16"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, k_1: DIMENSIONLESS, k_2: DIMENSIONLESS, height: M) -> None:
        """[:math:`l_{0}`] Effective length for unbraced members [:math:`m`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.3.2(3) - Formula (5.16)

        Parameters
        ----------
        k_1 : DIMENSIONLESS
            [:math:`k_{1}`] Relative flexibility of rotational constraint at end 1 [:math:`-`].
        k_2 : DIMENSIONLESS
            [:math:`k_{2}`] Relative flexibility of rotational constraint at end 2 [:math:`-`].
        height : M
            [:math:`l`] Clear height of compression member between end restraints [:math:`M`].
        """
        super().__init__()
        self.k_1 = k_1
        self.k_2 = k_2
        self.height = height

    @staticmethod
    def _evaluate(
        k_1: DIMENSIONLESS,
        k_2: DIMENSIONLESS,
        height: M,
    ) -> M:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(k_1=k_1, k_2=k_2, height=height)
        return height * max(math.sqrt(1.0 + 10 * (k_1 * k_2 / (k_1 + k_2))), (1.0 + k_1 / (1.0 + k_1)) * (1.0 + k_2 / (1.0 + k_2)))

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.16."""
        return LatexFormula(
            return_symbol=r"l_0",
            result=f"{self:.3f}",
            equation=r"l \cdot max\left\{"
            r"\sqrt{1+10 \cdot \frac{k_1 \cdot k_2}{k_1+k_2}}; "
            r"\left(1+\frac{k_1}{1 + k_1}\right) \cdot \left(1 + \frac{k_2}{1 + k_2}\right) \right\}",
            numeric_equation=rf"{self.height:.3f} \cdot max\left\{{"
            rf"\sqrt{{1+10 \cdot \frac{{{self.k_1:.3f} \cdot {self.k_2:.3f}}}{{{self.k_1:.3f}+{self.k_2:.3f}}}}}; "
            rf"\left(1+\frac{{{self.k_1:.3f}}}{{1 + {self.k_1:.3f}}}\right) \cdot "
            rf"\left(1 + \frac{{{self.k_2:.3f}}}{{1 + {self.k_2:.3f}}}\right) \right\}}",
            comparison_operator_label="=",
        )



"""Formula 5.37 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form5Dot37CreepFactor(Formula):
    r"""Class representing formula 5.37 for the calculation of the creep factor, :math:`K_{\phi}`."""

    label = "5.37"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ck: MPA,
        lambda_: DIMENSIONLESS,
        phi_ef: DIMENSIONLESS,
    ) -> None:
        r"""[:math:`K_{\phi}`] Creep factor [:math:`-`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.8.3(4) - Formula (5.37)

        Parameters
        ----------
        f_ck : MPA
            [:math:`f_{ck}`] Characteristic compressive strength of concrete [:math:`MPa`].
        lambda_ : DIMENSIONLESS
            [:math:`\lambda`] Slenderness ratio, see 5.8.3.1 [-].
        phi_ef : DIMENSIONLESS
            [:math:`\phi_{ef}`] Effective creep ratio, see 5.8.4 [-].
        """
        super().__init__()
        self.f_ck = f_ck
        self.lambda_ = lambda_
        self.phi_ef = phi_ef

    @staticmethod
    def _evaluate(
        f_ck: MPA,
        lambda_: float,
        phi_ef: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            f_ck=f_ck,
            lambda_=lambda_,
            phi_ef=phi_ef,
        )

        return max(1 + (0.35 + f_ck / 200 - lambda_ / 150) * phi_ef, 1)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.37."""
        return LatexFormula(
            return_symbol=r"k_{\phi}",
            result=f"{self:.3f}",
            equation=r"\max\left(1 + \left(0.35 + \frac{f_{ck}}{200} - \frac{\lambda}{150}\right) \cdot \phi_{ef}; 1\right)",
            numeric_equation=rf"\max\left(1 + \left(0.35 + \frac{{{self.f_ck:.3f}}}{{200}} - \frac{{{self.lambda_:.3f}}}{{150}}\right) "
            rf"\cdot {self.phi_ef:.3f}; 1\right)",
            comparison_operator_label="=",
            unit="-",
        )



"""Formula 6.76 from NEN-EN 1992-1-1+C2:2011: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form6Dot76DesignFatigueStrengthConcrete(Formula):
    """Class representing formula 6.76 for the design fatigue strength of concrete, :math:`f_{cd,fat}`."""

    label = "6.76"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        k_1: DIMENSIONLESS,
        beta_cc_t0: DIMENSIONLESS,
        f_cd: MPA,
        f_ck: MPA,
    ) -> None:
        """[:math:`f_{cd,fat}`] Design fatigue strength of concrete in :math:`MPa`.

        NEN-EN 1992-1-1+C2:2011 art. 6.8.7(1) - Formula (6.76)

        Parameters
        ----------
        k_1 : DIMENSIONLESS
            [:math:`k_{1}`] k1 factor (:math:`-`)
        beta_cc_t0 : DIMENSIONLESS
            [:math:`β_{cc}(t_0)`] Coefficient for concrete strength at first load application see (3.1.2 (6)) (:math:`-`).
            [:math:`t_0`] The time of the start of the cyclic loading in concrete in days.
        f_cd : MPA
            [:math:`f_{cd}`] Design strength of concrete (:math:`MPa`)
        f_ck : MPA
            [:math:`f_{ck}`] Characteristic strength of concrete (:math:`MPa`)

        """
        super().__init__()
        self.k_1 = k_1
        self.beta_cc_t0 = beta_cc_t0
        self.f_cd = f_cd
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        k_1: DIMENSIONLESS,
        beta_cc_t0: DIMENSIONLESS,
        f_cd: MPA,
        f_ck: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(k_1=k_1, beta_cc_t0=beta_cc_t0, f_cd=f_cd, f_ck=f_ck)

        return k_1 * beta_cc_t0 * f_cd * (1 - f_ck / 250)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 6.76."""
        return LatexFormula(
            return_symbol=r"f_{cd,fat}",
            result=f"{self:.3f}",
            equation=r"k_{1} \cdot β_{cc}(t_0) \cdot f_{cd} \cdot \left(1-\frac{f_{ck}}{250}\right)",
            numeric_equation=rf"{self.k_1:.3f} \cdot {self.beta_cc_t0:.3f} \cdot {self.f_cd:.3f} \cdot "
            rf"\left(1-\frac{{{self.f_ck:.3f}}}{{250}}\right)",
            comparison_operator_label="=",
        )



"""Formula 2.1b from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

from blueprints.codes.eurocode.nen_9997_1_c2_2017 import NEN_9997_1_C2_2017
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form2Dot1bRepresentativeValue(Formula):
    """Class representing formula 2.1b for the calculation of the representative value :math:`F_{rep}` of actions."""

    label = "2.1b"
    source_document = NEN_9997_1_C2_2017

    def __init__(self, psi: DIMENSIONLESS, f_k: float) -> None:
        """[:math:`F_{rep}`] Representative value of actions.

        NEN 9997-1+C2:2017 art.2.4.6.1(2) - Formula (2.1b)

        Parameters
        ----------
        psi : DIMENSIONLESS
            [:math:`Ψ`] factor for converting the characteristic value to the representative value [-].
        f_k : float
            [:math:`F_{k}`] Characteristic value of actions.
        """
        super().__init__()
        self.psi = psi
        self.f_k = f_k

    @staticmethod
    def _evaluate(
        psi: DIMENSIONLESS,
        f_k: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(psi=psi)
        return psi * f_k

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 2.1b."""
        return LatexFormula(
            return_symbol=r"F_{rep}",
            result=f"{self:.2f}",
            equation=r"\psi \cdot F_k",
            numeric_equation=rf"{self.psi:.2f} \cdot {self.f_k:.2f}",
            comparison_operator_label="=",
        )



"""Formula 3.10 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DAYS, MM, MM2


class Form3Dot10CoefficientAgeConcreteDryingShrinkage(Formula):
    """Class representing formula 3.10 for the calculation of the coefficient for drying shrinkage due to age."""

    label = "3.10"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        t: DAYS,
        t_s: DAYS,
        h_0: MM,
    ) -> None:
        """[βds(t,ts)] Coefficient for drying shrinkage due to age of concrete [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(6) - Formula (3.10)

        Parameters
        ----------
        t : DAYS
            [t] Age in days of the concrete at the considered moment [days].
        t_s : DAYS
            [t] Age in days of the concrete at the start of the drying shrinkage [days].
        h_0 : MM
            [h0] fictional thickness of cross-section [mm].
            = 2 * Ac / u
            Use your own implementation of this formula or use the SubForm3Dot10FictionalCrossSection class.

        Returns
        -------
        None
        """
        super().__init__()
        self.t = t
        self.t_s = t_s
        self.h_0 = h_0

    @staticmethod
    def _evaluate(
        t: DAYS,
        t_s: DAYS,
        h_0: MM,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if t <= 0:
            raise ValueError(f"Invalid t: {t}. t cannot be negative or zero")
        if t_s < 0:
            raise ValueError(f"Negative t_s: {t_s}. t_s cannot be negative")
        if t <= t_s:
            raise ValueError("Invalid t and t_s combination. t has to be larger than t_s")
        if h_0 <= 0:
            raise ValueError(f"Invalid h_0: {h_0}. h_0 cannot be negative or zero")
        return (t - t_s) / ((t - t_s) + 0.04 * np.sqrt(h_0**3))

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.10 formula."""
        return LatexFormula(
            return_symbol=r"\beta_{ds}(t,t_s)",
            result=f"{self:.3f}",
            equation=r"\frac{(t - t_s)}{(t - t_s) + 0.04 \sqrt{h_0^3}}",
            numeric_equation=rf"\frac{{({self.t:.2f} - {self.t_s:.2f})}}{{({self.t:.2f} - {self.t_s:.2f}) + 0.04 \sqrt{{{self.h_0:.2f}^3}}}}",
            comparison_operator_label="=",
        )


class SubForm3Dot10FictionalCrossSection(Formula):
    """Class representing sub-formula for formula 3.10 for the calculation of fictional thickness of the cross-section."""

    label = "3.10"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        a_c: MM2,
        u: MM,
    ) -> None:
        """[h0] Fictional thickness of the cross-section [mm].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(6) - h0

        Parameters
        ----------
        a_c : MM2
            [Ac] Area of the cross-section of the concrete [mm²].
        u : MM
            [u] Circumference of part that is subjected to drying [mm].
        """
        super().__init__()
        self.a_c = a_c
        self.u = u

    @staticmethod
    def _evaluate(
        a_c: MM2,
        u: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        if a_c <= 0:
            raise ValueError(f"Invalid a_c: {a_c}. a_c cannot be negative or zero")
        if u <= 0:
            raise ValueError(f"Invalid u: {u}. u cannot be negative or zero")
        return 2 * a_c / u

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.10 subformula."""
        return LatexFormula(
            return_symbol=r"h_0",
            result=f"{self:.2f}",
            equation=r"2 \cdot A_c / u",
            numeric_equation=rf"2 \cdot {self.a_c:.2f} / {self.u:.2f}",
            comparison_operator_label="=",
        )



"""Formula 5.11N from NEN-EN 1992-1-1 C2:2011: Chapter 5 Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_fraction
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot11nShearSlendernessCorrectionFactor(Formula):
    """Class representing formula 5.11N for the calculation of the shear slenderness correction factor [:math:`k_{λ}`]."""

    label = "5.11N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        lambda_factor: DIMENSIONLESS,
    ) -> None:
        """[:math:`k_{λ}`] Shear slenderness correction factor.

        NEN-EN 1992-1-1+C2:2011 art.5.6.3(4) - Formula (5.11N)

        Parameters
        ----------
        lambda_factor : DIMENSIONLESS
            [:math:`λ`] ratio of the distance between point of zero and maximum moment after redistribution and
        effective depth, d [:math:`-`]

        Use your own implementation for this value or use :class:`Form5Dot12nRatioDistancePointZeroAndMaxMoment`.
        """
        super().__init__()
        self.lambda_factor = lambda_factor

    @staticmethod
    def _evaluate(
        lambda_factor: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(lambda_factor=lambda_factor)
        return (lambda_factor / 3) ** 0.5

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.11N."""
        return LatexFormula(
            return_symbol=r"k_{λ}",
            result=f"{self:.3f}",
            equation=r"\sqrt{\frac{λ}{3}}",
            numeric_equation=rf"\sqrt{latex_fraction(numerator=self.lambda_factor, denominator=3)}",
            comparison_operator_label="=",
        )



"""Plotter for Reinforced Rectangular Cross-Sections."""

# ruff: noqa: PLR0913, F821
from typing import TypeVar

from matplotlib import patches as mplpatches
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from shapely import Point

from blueprints.structural_sections.concrete.rebar import Rebar

T = TypeVar("T", bound="RectangularReinforcedCrossSection")  # type: ignore[name-defined]

RCS_CROSS_SECTION_COLOR = (0.827, 0.827, 0.827)
STIRRUP_COLOR = (0.412, 0.412, 0.412)
REBAR_COLOR = (0.717, 0.255, 0.055)


class RectangularCrossSectionPlotter:
    """Plotter for Reinforced Rectangular Cross-Sections (RRCS)."""

    def __init__(
        self,
        cross_section: T,
    ) -> None:
        """Initialize the RRCSPlotter.

        Parameters
        ----------
        cross_section: RectangularReinforcedCrossSection
            Reinforced cross-section to plot.
        """
        self.cross_section = cross_section
        self.fig: plt.Figure | None = None
        self.axes: list[Axes] = []

    def plot(
        self,
        figsize: tuple[float, float] = (15.0, 8.0),
        title: str | None = None,
        font_size_title: float = 18.0,
        font_size_legend: float = 10.0,
        include_legend: bool = True,
        font_size_dimension: float = 12.0,
        custom_text_legend: str | None = None,
        custom_text_width: str | None = None,
        custom_text_height: str | None = None,
        offset_line_width: float = 1.25,
        offset_line_height: float = 1.2,
        center_line_style: dict[str, float | str] | None = None,
        show: bool = False,
        axes_i: int = 0,
    ) -> plt.Figure:
        """Plots the cross-section.

        Parameters
        ----------
        figsize: tuple[float, float]
            Size of the plot window.
        title: str
            Title of the plot.
        font_size_title: float
            Font size of the title.
        font_size_legend: float
            Font size of the legend.
        include_legend: bool
            Include legend in the plot.
        font_size_dimension: float
            Font size of the dimensions.
        custom_text_legend: str
            Custom text for the legend.
        custom_text_width: str
            Custom text for the width dimension. Replaces the width of the cross-section with the custom text.
        custom_text_height: str
            Custom text for the height dimension. Replaces the height of the cross-section with the custom text.
        offset_line_width: float
            Offset of the width line.
        offset_line_height: float
            Offset of the height line.
        center_line_style: dict[str, float | str] | None
            Style of the center lines. Check matplotlib documentation for more information (Annotation-arrowprops).
        show: bool
            Show the plot.
        axes_i: int
            Index of the axes to plot on. Default is 0.

        Returns
        -------
        plt.Figure
            Matplotlib figure.
        """
        self._start_plot(figsize=figsize)
        self._add_rectangle(axes_i=axes_i)
        self._add_center_lines(axes_i=axes_i, style=center_line_style)
        self._add_dimension_lines(
            axes_i=axes_i,
            font_size_dimension=font_size_dimension,
            custom_text_height=custom_text_height,
            custom_text_width=custom_text_width,
            offset_line_width=offset_line_width,
            offset_line_height=offset_line_height,
        )
        self._add_stirrups(axes_i=axes_i)
        self._add_longitudinal_rebars(axes_i=axes_i)

        # set limits and title
        self.axes[axes_i].axis("off")
        self.axes[axes_i].axis("equal")
        self.axes[axes_i].set_title(
            label=title or "",
            fontdict={"fontsize": font_size_title},
        )

        if include_legend:
            self._add_legend(
                axes_i=axes_i,
                font_size_legend=font_size_legend,
                custom_legend_text=custom_text_legend,
            )
        if show:
            plt.show()  # pragma: no cover
        assert self.fig is not None
        return self.fig

    def _start_plot(self, figsize: tuple[float, float] = (15.0, 8.0)) -> tuple[float, float]:
        """Starts the plot by initialising a matplotlib plot window of the given size.

        Parameters
        ----------
        figsize: tuple[float, float]
            Size of the plot window.
        """
        plt.close("all")
        self.fig = plt.figure(figsize=figsize)
        self.axes = [self.fig.add_subplot()]
        return self.fig.get_figwidth(), self.fig.get_figheight()

    def _add_rectangle(
        self,
        edge_color: str = "black",
        axes_i: int = 0,
    ) -> mplpatches.Rectangle:
        """Adds a rectangle to the plot.

        Parameters
        ----------
        edge_color: str
            Color of the edge of the rectangle. Use any matplotlib color.
        axes_i: int
            Index of the axes to plot on. Default is 0.
        """
        patch = mplpatches.Rectangle(
            xy=(-self.cross_section.width / 2, -self.cross_section.height / 2),
            width=self.cross_section.width,
            height=self.cross_section.height,
            edgecolor=edge_color,
            facecolor=RCS_CROSS_SECTION_COLOR,
            fill=True,
            lw=1,
        )
        self.axes[axes_i].add_patch(patch)
        return patch

    def _add_center_lines(self, axes_i: int = 0, style: dict[str, float | str] | None = None) -> None:
        """Adds center lines to the plot.

        Parameters
        ----------
        axes_i: int
            Index of the axes to plot on. Default is 0.
        style: dict[str, float]
            Style of the center lines. Check matplotlib documentation for more information (Annotation-arrowprops).
        """
        default_style = {"arrowstyle": "-", "linewidth": 0.8, "color": "gray", "linestyle": "dashdot"}
        if style:
            default_style.update(style)
        offset_center_line = 1.05
        self.axes[axes_i].annotate(
            text="z",
            xy=(0, (-self.cross_section.height / 2) * offset_center_line),
            xytext=(0, (self.cross_section.height / 2) * offset_center_line),
            arrowprops=default_style,
            verticalalignment="bottom",
            horizontalalignment="center",
        )
        self.axes[axes_i].annotate(
            text="y",
            xy=((self.cross_section.width / 2) * offset_center_line, 0),
            xytext=(-(self.cross_section.width / 2) * offset_center_line, 0),
            arrowprops=default_style,
            verticalalignment="center",
            horizontalalignment="right",
        )

    def _add_dimension_lines(
        self,
        axes_i: int = 0,
        style: mplpatches.ArrowStyle | None = None,
        offset_line_width: float = 1.25,
        offset_line_height: float = 1.2,
        custom_text_width: str | None = None,
        custom_text_height: str | None = None,
        font_size_dimension: float = 12.0,
    ) -> None:
        """Adds dimension lines to the plot.

        Parameters
        ----------
        axes_i: int
            Index of the axes to plot on. Default is 0.
        style: dict[str, float]
            Style of the dimension lines. Check matplotlib documentation for more information (Annotation-arrowprops).
        offset_line_width: float
            Offset of the width line.
        offset_line_height: float
            Offset of the height line.
        custom_text_width: str
            Custom text for the width dimension. Replaces the width of the cross-section with the custom text.
        custom_text_height: str
            Custom text for the height dimension. Replaces the height of the cross-section with the custom text.
        font_size_dimension: float
            Font size of the dimensions.
        """
        # add the width dimension line
        diameter_line_style = {
            "arrowstyle": style or mplpatches.ArrowStyle(stylename="<->", head_length=0.5, head_width=0.5),
        }
        offset_width = (-self.cross_section.height / 2) * offset_line_width
        self.axes[axes_i].annotate(
            text="",
            xy=(-self.cross_section.width / 2, offset_width),
            xytext=(self.cross_section.width / 2, offset_width),
            verticalalignment="center",
            horizontalalignment="center",
            arrowprops=diameter_line_style,
            annotation_clip=False,
        )
        self.axes[axes_i].text(
            s=custom_text_width or f"{self.cross_section.width:.0f} mm",
            x=0,
            y=offset_width,
            verticalalignment="bottom",
            horizontalalignment="center",
            fontsize=font_size_dimension,
        )

        # add the height dimension line
        offset_height = (-self.cross_section.width / 2) * offset_line_height
        self.axes[axes_i].annotate(
            text="",
            xy=(offset_height, self.cross_section.height / 2),
            xytext=(offset_height, -self.cross_section.height / 2),
            verticalalignment="center",
            horizontalalignment="center",
            arrowprops=diameter_line_style,
            rotation=90,
            annotation_clip=False,
        )
        self.axes[axes_i].text(
            s=custom_text_height or f"{self.cross_section.height:.0f} mm",
            x=offset_height,
            y=0,
            verticalalignment="center",
            horizontalalignment="right",
            fontsize=font_size_dimension,
            rotation=90,
        )

    def _add_stirrups(
        self,
        axes_i: int = 0,
    ) -> None:
        """Adds stirrups to the plot.

        Parameters
        ----------
        axes_i: int
            Index of the axes to plot on. Default is 0.
        """
        for stirrup in self.cross_section.stirrups:
            left_bottom = Point(stirrup.geometry.exterior.coords[0])  # left bottom point of the stirrup (center line)
            self.axes[axes_i].add_patch(
                mplpatches.Rectangle(
                    xy=(left_bottom.x - stirrup.diameter / 2, left_bottom.y - stirrup.diameter / 2),
                    width=stirrup.ctc_distance_legs + stirrup.diameter,
                    height=self.cross_section.height - self.cross_section.covers.upper - self.cross_section.covers.lower,
                    facecolor=STIRRUP_COLOR,
                    fill=True,
                )
            )
            self.axes[axes_i].add_patch(
                mplpatches.Rectangle(
                    xy=(left_bottom.x + stirrup.diameter / 2, left_bottom.y + stirrup.diameter / 2),
                    width=stirrup.ctc_distance_legs - stirrup.diameter,
                    height=self.cross_section.height - self.cross_section.covers.upper - self.cross_section.covers.lower - 2 * stirrup.diameter,
                    facecolor=RCS_CROSS_SECTION_COLOR,
                    fill=True,
                )
            )

    def _add_longitudinal_rebars(
        self,
        axes_i: int = 0,
    ) -> None:
        """Adds longitudinal rebars to the plot.

        Parameters
        ----------
        axes_i: int
            Index of the axes to plot on. Default is 0.
        """
        for rebar in self.cross_section.longitudinal_rebars:
            self.axes[axes_i].add_patch(
                mplpatches.Circle(
                    xy=(rebar.x, rebar.y),
                    radius=rebar.radius,
                    linewidth=1,
                    color=REBAR_COLOR,
                )
            )

    def legend_text(self) -> str:
        """Creates the legend text.

        Returns
        -------
        str
            Legend text.
        """
        # start building legend
        main_steel_material_used = self.cross_section.get_present_steel_materials()[0].name
        legend_text = f"{self.cross_section.concrete_material.concrete_class.value} - {main_steel_material_used}"

        legend_text += self._add_stirrups_to_legend()
        legend_text += self._add_longitudinal_rebars_to_legend()
        legend_text += self._add_rebar_configurations_to_legend()
        legend_text += self._add_single_longitudinal_rebars_to_legend()
        legend_text += self._add_covers_info_to_legend()

        return legend_text

    def _add_stirrups_to_legend(self) -> str:
        """Adds stirrups to the legend text."""
        stirrups_text = ""
        if self.cross_section.stirrups:
            stirrups_text += f"\nStirrups ({sum(stirrup.as_w for stirrup in self.cross_section.stirrups):.0f} mm²/m):"
            for stirrup in self.cross_section.stirrups:
                stirrups_text += (
                    f"\n  ⌀{stirrup.diameter}-{stirrup.distance} mm (b:{stirrup.ctc_distance_legs:.0f} mm) ({stirrup.as_w:.0f} " f"mm²/m)"
                )
        return stirrups_text

    def _add_longitudinal_rebars_to_legend(self) -> str:
        """Add longitudinal rebars to the legend text."""
        longitudinal_rebars = ""
        if self.cross_section.longitudinal_rebars:
            longitudinal_rebars += f"\nReinforcement ({sum(rebar.area for rebar in self.cross_section.longitudinal_rebars):.0f} mm²/m): "
        return longitudinal_rebars

    def _add_single_longitudinal_rebars_to_legend(self) -> str:
        """Add single longitudinal rebars to legend text."""
        single_longitudinal_text = ""
        if self.cross_section._single_longitudinal_rebars:  # noqa: SLF001
            rebar_diameters: dict[float, list[Rebar]] = {}
            for rebar in self.cross_section._single_longitudinal_rebars:  # noqa: SLF001
                rebar_diameters.setdefault(rebar.diameter, []).append(rebar)
            for diameter, rebars in rebar_diameters.items():
                single_longitudinal_text += f"\n  {len(rebars)}⌀{round(diameter, 2)} ({int(sum(rebar.area for rebar in rebars))} mm²/m)"
        return single_longitudinal_text

    def _add_rebar_configurations_to_legend(self) -> str:
        """Add rebar configurations to legend text (quantity in line)."""
        rebar_configurations_text = ""
        if self.cross_section._reinforcement_configurations:  # noqa: SLF001
            for _, configuration in self.cross_section._reinforcement_configurations:  # noqa: SLF001
                rebar_configurations_text += f"\n  {configuration!s} ({int(configuration.area)} mm²/m)"
        return rebar_configurations_text

    def _add_covers_info_to_legend(self) -> str:
        """Add covers info to legend text."""
        covers_text = ""
        if self.cross_section.stirrups or self.cross_section.longitudinal_rebars:
            covers_text += "\n" + self.cross_section.covers.get_covers_info()
        return covers_text

    def _add_legend(
        self,
        axes_i: int = 0,
        custom_legend_text: str | None = None,
        font_size_legend: float = 10.0,
        offset_center_line: float = 1.05,
    ) -> None:
        """Adds the legend to the plot.

        Parameters
        ----------
        axes_i: int
            Index of the axes to plot on. Default is 0.
        custom_legend_text: str
            Custom legend text.
        font_size_legend: float
            Font size of the legend.
        offset_center_line: float
            Offset of the center line.
        """
        legend_text = custom_legend_text or self.legend_text()
        self.axes[axes_i].annotate(
            text=legend_text,
            xy=((self.cross_section.width / 2) * offset_center_line, -self.cross_section.height / 2),
            verticalalignment="bottom",
            horizontalalignment="left",
            fontsize=font_size_legend,
            annotation_clip=False,
        )



"""Formula 3.17 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA


class Form3Dot17CompressiveStressConcrete(Formula):
    """Class representing formula 3.17 for the calculation of compressive stress in concrete using stress-strain diagram of figure 3.3."""

    label = "3.17"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_cd: MPA,
        epsilon_c: float,
        epsilon_c2: float,
        n: float,
    ) -> None:
        """[σc] Compressive stress in concrete using stress-strain diagram of figure 3.3 [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.7(1) - Formula (3.17)

        Parameters
        ----------
        f_cd : MPA
            [fcd] Design value compressive strength concrete [MPa].
        epsilon_c : float
            [εc] Strain in concrete [-].
        epsilon_c2 : float
            [εc2] Strain in concrete when reaching maximum strength following table 3.1 [-].
        n : float
            Exponent following table 3.1.

        Returns
        -------
        None
        """
        super().__init__()
        self.f_cd = f_cd
        self.epsilon_c = epsilon_c
        self.epsilon_c2 = epsilon_c2
        self.n = n

    @staticmethod
    def _evaluate(
        f_cd: MPA,
        epsilon_c: float,
        epsilon_c2: float,
        n: float,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_cd < 0:
            raise ValueError(f"Invalid f_cd: {f_cd}. f_cd cannot be negative")
        if epsilon_c < 0:
            raise ValueError(f"Invalid epsilon_c: {epsilon_c}. epsilon_c cannot be negative")
        if epsilon_c > epsilon_c2:
            raise ValueError(f"epsilon_c: {epsilon_c} > epsilon_c2: {epsilon_c2}. Try using Form3Dot18CompressiveStressConcrete class.")
        return f_cd * (1 - (1 - (epsilon_c / epsilon_c2)) ** n)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.17."""
        return LatexFormula(
            return_symbol=r"\sigma_c",
            result=f"{self:.3f}",
            equation=r"f_{cd} \cdot \left[ 1 - \left( 1 - \frac{\epsilon_c}{\epsilon_{c2}} \right)^n \right]",
            numeric_equation=(
                rf"{self.f_cd:.3f} \cdot \left[ 1 - \left( 1 - \frac{{{self.epsilon_c:.3f}}}"
                rf"{{{self.epsilon_c2:.3f}}} \right)^{{{self.n:.2f}}} \right]"
            ),
            comparison_operator_label="=",
        )



"""Formula 8.21 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot21AnchorageLength(Formula):
    """Class representing formula 8.21 for the calculation of anchorage length :math:`l_{bpd}` [:math:`mm`].

    NEN-EN 1992-1-1+C2:2011 art.8.10.2.3(4) - Formula (8.21)

    Parameters
    ----------
    l_pt2 : MM
        [:math:`l_{pt2}`] is the upper design value of transmission length, see 8.10.2.2 (3) [:math:`mm`].
    alpha_2 : DIMENSIONLESS
        [:math:`α_{2}`] as defined in 8.10.2.2 (2) [-].
    diameter : MM
        [:math:`Ø`] Diameter of the tendon [:math:`mm`].
    sigma_pd : MPA
        [:math:`σ_{pd}`] Is the tendon stress corresponding to the force described in art.8.10.2.3(1) [:math:`MPa`].
    sigma_pminf : MPA
        [:math:`σ_{pm∞}`] is the prestress after all losses [:math:`MPa`].
    f_bpd : MPA
       [:math:`f_{bpd}`] Bond strength for anchorage in the ultimate limit state [:math:`MPa`].
    """

    label = "8.21"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        l_pt2: MM,
        alpha_2: DIMENSIONLESS,
        diameter: MM,
        sigma_pd: MPA,
        sigma_pminf: MPA,
        f_bpd: MPA,
    ) -> None:
        super().__init__()
        self.l_pt2 = l_pt2
        self.alpha_2 = alpha_2
        self.diameter = diameter
        self.sigma_pd = sigma_pd
        self.sigma_pminf = sigma_pminf
        self.f_bpd = f_bpd

    @staticmethod
    def _evaluate(l_pt2: MM, alpha_2: DIMENSIONLESS, diameter: MM, sigma_pd: MPA, sigma_pminf: MPA, f_bpd: MPA) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(l_pt2=l_pt2, alpha_2=alpha_2, diameter=diameter, sigma_pd=sigma_pd, sigma_pminf=sigma_pminf)
        raise_if_less_or_equal_to_zero(f_bpd=f_bpd)
        return l_pt2 + alpha_2 * diameter * (sigma_pd - sigma_pminf) / f_bpd

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 8.21."""
        return LatexFormula(
            return_symbol=r"l_{bpd}",
            result=f"{self:.3f}",
            equation=r"l_{pt2} + \alpha_{2} \cdot Ø \cdot \frac{\sigma_{pd} - \sigma_{pm\infty}}{f_{bpd}}",
            numeric_equation=rf"{self.l_pt2:.3f} + {self.alpha_2:.3f} \cdot {self.diameter:.3f} \cdot \frac{{{self.sigma_pd:.3f} - "
            rf"{self.sigma_pminf:.3f}}}{{{self.f_bpd:.3f}}}",
            comparison_operator_label="=",
        )



"""Formula 3.12 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA


class Form3Dot12AutogeneShrinkageInfinity(Formula):
    """Class representing formula 3.12, which calculates the autogene shrinkage at infinity."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "3.12"

    def __init__(
        self,
        f_ck: MPA,
    ) -> None:
        """[εca(∞)] Autogene shrinkage at infinity [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(6) - Formula (3.12)

        Parameters
        ----------
        f_ck : MPA
            [fck] Compressive strength concrete [MPa].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        f_ck: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_ck < 0:
            raise ValueError(f"Invalid f_ck: {f_ck}. f_ck cannot be negative")
        return 2.5 * (f_ck - 10) * 10**-6

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.12."""
        return LatexFormula(
            return_symbol=r"\epsilon_{ca}(\infty)",
            result=f"{self:.6f}",
            equation=r"2.5 \cdot (f_{ck} - 10) \cdot 10^{-6}",
            numeric_equation=rf"2.5 \cdot ({self.f_ck:.3f} - 10) \cdot 10^{{-6}}",
            comparison_operator_label="=",
        )



"""Formula 9.13 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailling and specific rules."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, MM
from blueprints.validations import raise_if_negative


class Form9Dot13TensileForceToBeAnchored(Formula):
    """Class representing the formula 9.13 for the calculation of the tensile force to be anchored for."""

    label = "9.13"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        r: MM,
        z_e: MM,
        z_i: MM,
    ) -> None:
        """[Fs] Tensile force to be anchored [kN].

        NEN-EN 1992-1-1+C2:2011 art.9.8.2.2(2) - Formula (9.13)

        Parameters
        ----------
        r: MM
            [R] The resultant of ground pressure within x from figure 9.13 [mm].
        z_e: MM
            [ze] The external lever arm, see figure 9.13, i.e. distance between the reinforcement and the horizontal force Fc [mm].
        z_i: MM
            [zi] Internal lever arm, see figure 9.13, i.e. distance between R and the vertical force NEd [mm].
        """
        super().__init__()
        self.r = r
        self.z_e = z_e
        self.z_i = z_i

    @staticmethod
    def _evaluate(
        r: MM,
        z_e: MM,
        z_i: MM,
    ) -> KN:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(r=r, z_e=z_e, z_i=z_i)
        return r * z_e / z_i

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 9.13."""
        return LatexFormula(
            return_symbol=r"F_s",
            result=f"{self:.2f}",
            equation=r"R \cdot z_e / z_i",
            numeric_equation=rf"{self.r:.2f} \cdot {self.z_e:.2f} / {self.z_i:.2f}",
            comparison_operator_label="=",
        )



"""Formula 5.31 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KNM


class Form5Dot31DesignMoment(Formula):
    """Class representing formula 5.31 for the calculation of the design moment, :math:`M_{Ed}`."""

    label = "5.31"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        m_0ed: KNM,
        m_2: KNM,
    ) -> None:
        """[:math:`M_{Ed}`] Design moment [:math:`kNm`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.8.2(2) - Formula (5.31)

        Parameters
        ----------
        m_0ed : KNM
            [:math:`M_{0Ed}`] First order moment, including the effect of imperfections; see also 5.8.8.2 (2) [:math:`kNm`].
        m_2 : KNM
            [:math:`M_{2}`] Nominal 2nd order moment; see 5.8.8.2 (3) [:math:`kNm`].
        """
        super().__init__()
        self.m_0ed = m_0ed
        self.m_2 = m_2

    @staticmethod
    def _evaluate(
        m_0ed: KNM,
        m_2: KNM,
    ) -> KNM:
        """Evaluates the formula, for more information see the __init__ method."""
        return m_0ed + m_2

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.31."""
        return LatexFormula(
            return_symbol=r"M_{Ed}",
            result=f"{self:.3f}",
            equation=r"M_{0Ed} + M_{2}",
            numeric_equation=rf"{self.m_0ed:.3f} + {self.m_2:.3f}",
            comparison_operator_label="=",
            unit="kNm",
        )



"""Custom metaclass for ABCEnum."""

from abc import ABCMeta
from enum import EnumMeta


class ABCEnumMeta(ABCMeta, EnumMeta):
    """Custom metaclass for ABCEnum.

    This metaclass is a combination of ABCMeta and EnumMeta.
    It is used to check if abstract methods are implemented in the class.
    """

    def __new__(cls, *args, **kwarg) -> type:
        """Create a new instance of the class.

        Here we check if the class has abstract methods that are not implemented.
        If so, we raise a TypeError.
        """
        abstract_enum_cls = super().__new__(cls, *args, **kwarg)
        # Enum classes have a _member_map_ attribute that is a dictionary of the enum members.
        # If the class has abstract methods and the _member_map_ attribute is not empty, we check if the abstract methods are implemented.
        if getattr(abstract_enum_cls, "_member_map_", False) and getattr(abstract_enum_cls, "__abstractmethods__", False):
            abstract_methods = list(abstract_enum_cls.__abstractmethods__)
            missing = ", ".join(f"{method!r}" for method in abstract_methods)
            plural = "s" if len(abstract_methods) > 1 else ""
            raise TypeError(
                f"Can't instantiate abstract class {abstract_enum_cls.__name__!r}" f" without an implementation for abstract method{plural} {missing}"
            )
        return abstract_enum_cls



"""Formula 5.27 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form5Dot27EffectiveDesignModulusElasticity(Formula):
    """Class representing formula 5.27 for the calculation of the effective design modulus of elasticity, :math:`E_{cd,eff}`."""

    label = "5.27"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, e_cd: MPA, phi_ef: DIMENSIONLESS) -> None:
        r"""[:math:`E_{cd,eff}`] Effective design modulus of elasticity.

        NEN-EN 1992-1-1+C2:2011 art.5.8.6(3) - Formula (5.27)

        Parameters
        ----------
        e_cd : MPA
            [:math:`E_{cd}`] is the design modulus of elasticity of concrete according to 5.8.6 (3).
        phi_ef : DIMENSIONLESS
            [:math:`\phi_{ef}`] is the effective creep coefficient; same value as for columns may be used.
        """
        super().__init__()
        self.e_cd = e_cd
        self.phi_ef = phi_ef

    @staticmethod
    def _evaluate(e_cd: MPA, phi_ef: DIMENSIONLESS) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(e_cd=e_cd, phi_ef=phi_ef)
        return e_cd / (1 + phi_ef)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.27."""
        return LatexFormula(
            return_symbol=r"E_{cd,eff}",
            result=f"{self:.3f}",
            equation=r"\frac{E_{cd}}{1 + \phi_{ef}}",
            numeric_equation=rf"\frac{{{self.e_cd:.3f}}}{{1 + {self.phi_ef:.3f}}}",
            comparison_operator_label="=",
        )



"""Formula 8.5 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import RATIO
from blueprints.validations import raise_if_negative


class Form8Dot5ProductAlphas235(Formula):
    """Class representing formula 8.5 for the calculating the product of α_{2}, α_{3} α_{5} [-]."""

    label = "8.5"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        alpha_2: RATIO,
        alpha_3: RATIO,
        alpha_5: RATIO,
    ) -> None:
        """Calculate the product of α_{2}, α_{3} and α_{5} [:math:`-`].

        NEN-EN 1992-1-1+C2:2011 art.8.4.4(1) - Formula (8.5) prescribes that :math:`(α_{2} ⋅ α_{3} ⋅ α_{5}) ≥ 0.7`.
        Used by NEN-EN 1992-1-1+C2:2011 art.8.4.4(1) - Formula (8.4)

        Parameters
        ----------
        alpha_2 : RATIO
            [:math:`α_{2}`] Coefficient for the effect of minimum concrete cover (see figure 8.3) [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1 - 0.15 ⋅ (c_{d} - Ø) / Ø <= 1` with a minimum of 0.7 for straight bars in tension.

            :math:`= 1 - 0.15 ⋅ (c_{d} - 3 ⋅ Ø) / Ø <= 1` with a minimum of 0.7 for bars other than
            straight in tension (see figure 8.1 (b), (c) and (d)).

            Note: see figure 8.3 for values of :math:`c_{d}`.
        alpha_3 : RATIO
            [:math:`α_{3}`] Coefficient for the effect of confinement by transverse reinforcement [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1 - K ⋅ λ <= 1` with a minimum of 0.7 for bars in tension.

            Where: :math:`λ = (ΣA_{st} - ΣA_{st,min}) / A_{s}`.

            Where: :math:`ΣA_{st,min}` = cross-sectional area of the minimum transverse
            reinforcement :math:`= 0,25 ⋅ A_{s}` for beams and 0 for slabs.

            Note: see figure 8.4 for values of :math:`K, A_{s} and A_{st}`.

        alpha_5 : RATIO
            [:math:`α_{5}`] Coefficient for the effect of the pressure transverse to the plane of splitting
            along the design anchorage length :math:`l_{bd}` (see 8.6) [-].

            :math:`= 1 - 0.04 ⋅ p <= 1` with a minimum of 0.7 for all types of anchorage in compression.

            Where: p = transverse pressure at ultimate limit state along :math:`l_{bd}` [MPa].
        """
        super().__init__()
        self.alpha_2 = alpha_2
        self.alpha_3 = alpha_3
        self.alpha_5 = alpha_5

    @staticmethod
    def _evaluate(
        alpha_2: RATIO,
        alpha_3: RATIO,
        alpha_5: RATIO,
    ) -> RATIO:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
        )

        # The product of alpha values 2, 3, 5 can't be smaller than 0.7
        return max(alpha_2 * alpha_3 * alpha_5, 0.7)

    def latex(self) -> LatexFormula:
        """Returns a LatexFormula representation of the formula."""
        return LatexFormula(
            return_symbol="",
            result=f"{self:.2f}",
            equation=r"\alpha_2 \cdot \alpha_3 \cdot \alpha_5 \ge 0.7",
            numeric_equation=rf"{self.alpha_2} \cdot {self.alpha_3} \cdot {self.alpha_5} \ge 0.7",
            comparison_operator_label="\\to",
        )



"""Formula 5.38a from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot38aCheckRelativeSlendernessRatio(Formula):
    r"""Class representing formula 5.38a for check of relative slenderness ratio."""

    label = "5.38a"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        lambda_y: DIMENSIONLESS,
        lambda_z: DIMENSIONLESS,
    ) -> None:
        r"""Check the ratio of the slenderness in y-direction and z-direction.

        NEN-EN 1992-1-1+C2:2011 art.5.8.9(3) - Formula (5.38a)

        Parameters
        ----------
        lambda_y : DIMENSIONLESS
            [:math:`\lambda_{y}`] Slenderness ratio in y-direction [-].
        lambda_z : DIMENSIONLESS
            [:math:`\lambda_{z}`] Slenderness ratio in z-direction [-].
        """
        super().__init__()
        self.lambda_y = lambda_y
        self.lambda_z = lambda_z

    @staticmethod
    def _evaluate(
        lambda_y: float,
        lambda_z: float,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(lambda_y=lambda_y, lambda_z=lambda_z)

        return (lambda_y / lambda_z <= 2) and (lambda_z / lambda_y <= 2)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.38a."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"\left( \frac{\lambda_{y}}{\lambda_{z}} \leq 2 \text{ and } \frac{\lambda_{z}}{\lambda_{y}} \leq 2 \right)",
            numeric_equation=rf"\left( \frac{{{self.lambda_y:.3f}}}{{{self.lambda_z:.3f}}} \leq 2 \text{{ and }} "
            rf"\frac{{{self.lambda_z:.3f}}}{{{self.lambda_y:.3f}}} \leq 2 \right)",
            comparison_operator_label="\\to",
            unit="",
        )



"""Formula 8.1 from NEN-EN 1992-1-1+C2:2011: Chapter 8 Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, MM, MPA
from blueprints.unit_conversion import KN_TO_N
from blueprints.validations import raise_if_negative


class Form8Dot1RequiredMinimumMandrelDiameter(Formula):
    """Class representing formula 8.1 for the calculation of the required minimum mandrel diameter if it needs to be checked to avoid
    concrete failure.
    """

    label = "8.1"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_bt: KN,
        a_b: MM,
        diameter: MM,
        f_cd: MPA,
    ) -> None:
        """[Øm,min] minimum mandrel diameter if it needs to be checked to avoid concrete failure [MM].

        NEN-EN 1992-1-1+C2:2011 art.8.3(3) - Formula (8.1)

        Parameters
        ----------
        f_bt: KN
            [Fbt] Tensile force from ultimate loads in a bar or group of bars in contact at the start of a bend  [kN].
        a_b: MM
            [ab] Half of the centre-to-centre distance between bars (or groups of bars) perpendicular
            to the plane of the bend for a given bar (or group of bars in contact).
            For a bar or group of bars adjacent to the face of the member, 'ab' should be taken as the cover plus Ø/2 [mm].
        diameter: MM
            [Ø] Diameter of reinforcing bar [mm].
        f_cd: MPA
            [fcd] Design value of concrete compressive stress [MPa].
            Note: The value of fcd should not be taken greater than that for concrete class C55/67.
        """
        super().__init__()
        self.f_bt = f_bt
        self.a_b = a_b
        self.diameter = diameter
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        f_bt: KN,
        a_b: MM,
        diameter: MM,
        f_cd: MPA,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            f_bt=f_bt,
            a_b=a_b,
            diameter=diameter,
            f_cd=f_cd,
        )
        return f_bt * KN_TO_N * ((1 / a_b) + 1 / (2 * diameter)) / f_cd

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 8.1."""
        return LatexFormula(
            return_symbol=r"Ø_{m,min}",
            result=f"{self:.2f}",
            equation=r"\frac{F_{bt} \left( \frac{1}{a_b} + \frac{1}{2 \cdot Ø} \right) }{f_{cd}}",
            numeric_equation=(
                rf"\frac{{{self.f_bt:.2f} \cdot 1000 \cdot \left( \frac{{1}}{{{self.a_b:.2f}}}"
                rf" + \frac{{1}}{{2 \cdot {self.diameter:.2f}}} \right)}}{{{self.f_cd:.2f}}}"
            ),
            comparison_operator_label="=",
        )



"""Formula 3.9 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula


class Form3Dot9DryingShrinkage(Formula):
    """Class representing formula 3.9 for the calculation of the drying shrinkage."""

    label = "3.9"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        beta_ds_tt_s: float,
        k_h: float,
        epsilon_cd_0: float,
    ) -> None:
        """[εcd(t)] Development of the drying shrinkage [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(6) - Formula (3.9)

        Parameters
        ----------
        beta_ds_tt_s : float
            [βds(t, ts)] Coefficient that depends on the age t (in days) of the concrete for the drying shrinkage [-].
        k_h : float
            [kh] Coefficient depending on the fictional thickness h0 following table 3.3 [-].
            h0 = 100 -> kh = 1.0
            h0 = 200 -> kh = 0.85
            h0 = 300 -> kh = 0.75
            h0 >= 500 -> kh = 0.70
        epsilon_cd_0 : float
            [εcd,0] Nominal unobstructed drying shrinkage, formula in appendix B or use table 3.2 [-].

        Returns
        -------
        None
        """
        super().__init__()
        self.beta_ds_tt_s = beta_ds_tt_s
        self.k_h = k_h
        self.epsilon_cd_0 = epsilon_cd_0

    @staticmethod
    def _evaluate(
        beta_ds_tt_s: float,
        k_h: float,
        epsilon_cd_0: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if beta_ds_tt_s < 0:
            raise ValueError(f"Negative beta_ds_tt_s: {beta_ds_tt_s}. beta_ds_tt_s cannot be negative")
        if k_h < 0:
            raise ValueError(f"Negative k_h: {k_h}. k_h cannot be negative")
        return beta_ds_tt_s * k_h * epsilon_cd_0

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.9."""
        return LatexFormula(
            return_symbol=r"\epsilon_{cd}(t)",
            result=f"{self:.3f}",
            equation=r"\beta_{ds}(t,t_s) \cdot k_h \cdot \epsilon_{cd,0}",
            numeric_equation=rf"{self.beta_ds_tt_s:.3f} \cdot {self.k_h:.3f} \cdot {self.epsilon_cd_0:.3f}",
            comparison_operator_label="=",
        )



"""Formula 5.26 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot26FactorKs(Formula):
    """Class representing the factor K_s that represents the factor for contribution of reinforcement to the resistance."""

    label = "5.26"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, rho: float) -> None:
        r"""[:math:`K_s`] Factor K_s = 0.

        NEN-EN 1992-1-1+C2:2011 art.5.8.7.2(2) - Formula (5.26)

        Parameters
        ----------
        rho : float
            [:math:`\rho`] Geometric reinforcement ratio, As/Ac. Must be >= 0.01.
        """
        super().__init__()
        self.rho = rho

    @staticmethod
    def _evaluate(rho: float) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if rho <= 0.01:
            raise ValueError(f"Invalid rho: {rho}. rho cannot be less than 0.01")
        return 0.0

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.26 K_s."""
        return LatexFormula(
            return_symbol=r"K_s",
            result=f"{self._evaluate(self.rho):.3f}",
            equation=r"0",
            numeric_equation="0",
            comparison_operator_label="=",
        )


class Form5Dot26FactorKc(Formula):
    """Class representing the factor K_c representing the factor for effects of cracking, creep, etc."""

    label = "5.26"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, phi_ef: DIMENSIONLESS, rho: float) -> None:
        r"""[:math:`K_c`] Factor K_c = 0.3 / (1 + 0.5 * \phi_{ef}).

        NEN-EN 1992-1-1+C2:2011 art.5.8.7.2(2) - Formula (5.26)

        Parameters
        ----------
        phi_ef : DIMENSIONLESS
            [:math:`\phi_{ef}`] Effective creep ratio, see 5.8.4.
        rho : float
            [:math:`\rho`] Geometric reinforcement ratio, As/Ac. Must be > 0.01.
        """
        super().__init__()
        self.phi_ef = phi_ef
        self.rho = rho

    @staticmethod
    def _evaluate(phi_ef: DIMENSIONLESS, rho: float) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        if rho <= 0.01:
            raise ValueError(f"Invalid rho: {rho}. rho must be greater than 0.01")
        raise_if_less_or_equal_to_zero(phi_ef=phi_ef)
        return 0.3 / (1 + 0.5 * phi_ef)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.26 K_c."""
        return LatexFormula(
            return_symbol=r"K_c",
            result=f"{self:.3f}",
            equation=r"\frac{0.3}{1 + 0.5 \cdot \phi_{ef}}",
            numeric_equation=rf"\frac{{{0.3}}}{{1 + 0.5 \cdot {self.phi_ef:.3f}}}",
            comparison_operator_label="=",
        )



"""Formula 5.32 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KNM


class Form5Dot32EquivalentFirstOrderEndMoment(Formula):
    """Class representing formula 5.32 for the calculation of the equivalent first order end moment, :math:`M_{0e}`."""

    label = "5.32"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        m_01: KNM,
        m_02: KNM,
    ) -> None:
        """[:math:`M_{0e}`] Equivalent first order end moment [:math:`kNm`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.8.2(2) - Formula (5.32)

        Parameters
        ----------
        m_01 : KNM
            [:math:`M_{01}`] The smaller first order end moment [:math:`kNm`].
        m_02 : KNM
            [:math:`M_{02}`] The larger first order end moment [:math:`kNm`].
        """
        super().__init__()
        self.m_01 = m_01
        self.m_02 = m_02

    @staticmethod
    def _evaluate(
        m_01: KNM,
        m_02: KNM,
    ) -> KNM:
        """Evaluates the formula, for more information see the __init__ method."""
        if abs(m_02) < abs(m_01):
            raise ValueError("The absolute value of M_02 must be equal to or larger than the absolute value of M_01.")

        return max(0.6 * m_02 + 0.4 * m_01, 0.4 * m_02)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.32."""
        return LatexFormula(
            return_symbol=r"M_{0e}",
            result=f"{self:.3f}",
            equation=r"\max\left(0.6 \cdot M_{02} + 0.4 \cdot M_{01}; 0.4 \cdot M_{02}\right)",
            numeric_equation=rf"\max\left(0.6 \cdot {self.m_02:.3f} + 0.4 \cdot {self.m_01:.3f}; 0.4 \cdot {self.m_02:.3f}\right)",
            comparison_operator_label="=",
            unit="kNm",
        )



"""Formula 3.11 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula


class Form3Dot11AutogeneShrinkage(Formula):
    """Class representing formula 3.11, which calculates the autogene shrinkage."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "3.11"

    def __init__(
        self,
        beta_as_t: float,
        epsilon_ca_inf: float,
    ) -> None:
        """[εca(t)] Autogene shrinkage [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(6) - Formula (3.11)

        Parameters
        ----------
        beta_as_t : float
            [βas(t)] Coefficient dependent on time in days for autogene shrinkage [-].
            = 1 - exp(-0.2 * t^0.5)
            Use your own implementation of this formula or use the Form3Dot13CoefficientTimeAutogeneShrinkage class
        epsilon_ca_inf : float
            [εca(∞)] Autogene shrinkage at infinity [-].
            = 2.5 * (fck - 10) E-6
            Use your own implementation of this formula or use the Form3Dot12AutogeneShrinkageInfinity class.

        Returns
        -------
        None
        """
        super().__init__()
        self.beta_as_t = beta_as_t
        self.epsilon_ca_inf = epsilon_ca_inf

    @staticmethod
    def _evaluate(
        beta_as_t: float,
        epsilon_ca_inf: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if beta_as_t < 0:
            raise ValueError(f"Invalid beta_as_t: {beta_as_t}. beta_as_t cannot be negative")
        return beta_as_t * epsilon_ca_inf

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.11."""
        return LatexFormula(
            return_symbol=r"\epsilon_{ca}(t)",
            result=f"{self:.3f}",
            equation=r"\beta_{as}(t) \cdot \epsilon_{ca}(\infty)",
            numeric_equation=rf"{self.beta_as_t:.3f} \cdot {self.epsilon_ca_inf:.3f}",
            comparison_operator_label="=",
        )



"""Formula 5.23 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form5Dot23FactorConcreteStrengthClass(Formula):
    """Class representing formula 5.23 for the calculation of the factor for concrete strength class, :math:`k_{1}`."""

    label = "5.23"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, f_ck: MPA) -> None:
        r"""[:math:`k_{1}`] Factor for concrete strength class.

        NEN-EN 1992-1-1+C2:2011 art.5.8.6(3) - Formula (5.23)

        Parameters
        ----------
        f_ck : MPA
            [:math:`f_{ck}`] is the characteristic compressive strength of concrete.
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(f_ck: MPA) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(f_ck=f_ck)
        return (f_ck / 20) ** 0.5

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.23."""
        return LatexFormula(
            return_symbol=r"k_{1}",
            result=f"{self:.3f}",
            equation=r"\sqrt{\frac{f_{ck}}{20}}",
            numeric_equation=rf"\sqrt{{\frac{{{self.f_ck:.3f}}}{{20}}}}",
            comparison_operator_label="=",
        )



"""Formula 8.18 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form8Dot18DesignValueTransmissionLength2(Formula):
    """Class representing formula 8.18 for the calculation of design value 2 of the transmission length :math:`l_{pt2}`. The less favourable of
    :math:`l_{pt1}` or :math:`l_{pt2}` has to be chosen depending on the design situation.
    """

    label = "8.18"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, l_pt: MM) -> None:
        """[:math:`l_{pt2}`] design value 2 of the transmission length [:math:`mm`].

        NEN-EN 1992-1-1+C2:2011 art.8.10.2.2(3) - Formula (8.18)

        Parameters
        ----------
        l_pt : MM
            [:math:`l_{pt}`] Basic value of the transmission length [:math:`mm`].
            Use your own implementation for this value or use :class:`Form8Dot16BasicTransmissionLength` class.
        """
        super().__init__()
        self.l_pt = l_pt

    @staticmethod
    def _evaluate(
        l_pt: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(l_pt=l_pt)
        return 1.2 * l_pt

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 8.18."""
        return LatexFormula(
            return_symbol=r"l_{pt2}",
            result=f"{self:.2f}",
            equation=r"1.2 \cdot l_{pt}",
            numeric_equation=rf"1.2 \cdot {self.l_pt:.2f}",
            comparison_operator_label="=",
        )



"""Formula 5.15 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

import math

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_negative


class Form5Dot15EffectiveLengthBraced(Formula):
    """Class representing formula 5.15 for the calculation of the effective length of braced members, :math:`l_0`."""

    label = "5.15"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, k_1: DIMENSIONLESS, k_2: DIMENSIONLESS, height: M) -> None:
        """[:math:`l_{0}`] Effective length for braced members [:math:`m`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.3.2(3) - Formula (5.15)

        Parameters
        ----------
        k_1 : DIMENSIONLESS
            [:math:`k_{1}`] Relative flexibility of rotational constraint at end 1 [:math:`-`].
        k_2 : DIMENSIONLESS
            [:math:`k_{2}`] Relative flexibility of rotational constraint at end 2 [:math:`-`].
        height : M
            [:math:`l`] Clear height of compression member between end restraints [:math:`m`].
        """
        super().__init__()
        self.k_1 = k_1
        self.k_2 = k_2
        self.height = height

    @staticmethod
    def _evaluate(
        k_1: DIMENSIONLESS,
        k_2: DIMENSIONLESS,
        height: M,
    ) -> M:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(k_1=k_1, k_2=k_2, height=height)
        return 0.5 * height * math.sqrt((1 + k_1 / (0.45 + k_1)) * (1 + k_2 / (0.45 + k_2)))

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.15."""
        return LatexFormula(
            return_symbol=r"l_0",
            result=f"{self:.3f}",
            equation=r"0.5 \cdot l \cdot \sqrt{\left(1+\frac{k_1}{0.45 + k_1}\right) \cdot \left(1 + \frac{k_2}{0.45 + k_2}\right)}",
            numeric_equation=rf"0.5 \cdot {self.height:.3f} \cdot \sqrt{{\left(1+\frac{{{self.k_1:.3f}}}{{0.45 + "
            rf"{self.k_1:.3f}}}\right) \cdot \left(1 + \frac{{{self.k_2:.3f}}}{{0.45 + {self.k_2:.3f}}}\right)}}",
            comparison_operator_label="=",
        )



"""Formula A.1 from NEN-EN 1993-1-9+C2:2012: Annex A - Determination of fatigue load parameters and verification formats."""

from blueprints.codes.eurocode.nen_en_1993_1_9_c2_2012 import NEN_EN_1993_1_9_C2_2012
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_lists_differ_in_length, raise_if_negative


class FormADot1DamageDuringDesignLife(Formula):
    """Class representing formula A.1 for the calculation of the damage during the design life :math:`D_d`."""

    label = "A.1"
    source_document = NEN_EN_1993_1_9_C2_2012

    def __init__(self, n_e: list[DIMENSIONLESS], n_r: list[DIMENSIONLESS]) -> None:
        """[:math:`D_d`] The calculation of the damage during the design life [:math:`-`].

        NEN-EN 1993-1-9+C2:2012 art.A.5 - Formula (A.1)

        Parameters
        ----------
        n_e : list[DIMENSIONLESS]
            [:math:`n_E`] Contains number of cycles associated with the stress range γ_Ff·Δσ_i for each band i in the factored spectrum [:math:`-`].
        n_r : list[DIMENSIONLESS]
            [:math:`N_R`] Contains the endurance (in cycles) obtained from the factored Δσ_C/γ_Mf-n_r curve for each stress range of
            γ_Ff·Δ_i [:math:`-`]

        Returns
        -------
        None
        """
        super().__init__()
        self.n_e = n_e
        self.n_r = n_r

    @staticmethod
    def _evaluate(
        n_e: list[DIMENSIONLESS],
        n_r: list[DIMENSIONLESS],
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_lists_differ_in_length(n_e=n_e, n_r=n_r)
        raise_if_negative(n_e_min=min(n_e))
        raise_if_less_or_equal_to_zero(n_r_min=min(n_r))
        return sum([n_e[i] / n_r[i] for i in range(len(n_e))])

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula A.1."""
        return LatexFormula(
            return_symbol=r"D_d",
            result=f"{self:.3f}",
            equation=r"\sum_{i}^{n} \frac{n_{Ei}}{N_Ri}",
            numeric_equation="".join(rf"\frac{{{self.n_e[i]:.3f}}}{{{self.n_r[i]:.3f}}} + " for i in range(len(self.n_r)))[:-3],
            comparison_operator_label="=",
        )



"""Formula 5.28 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN, KNM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot28TotalDesignMoment(Formula):
    """Class representing formula 5.28 for the calculation of the total design moment, :math:`M_{Ed}`."""

    label = "5.28"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        m_0ed: KN,
        beta: DIMENSIONLESS,
        n_ed: KN,
        n_b: KN,
    ) -> None:
        """[:math:`M_{Ed}`] Total design moment [:math:`kNm`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.8.2(2) - Formula (5.28)

        Parameters
        ----------
        m_0ed : KNM
            [:math:`M_{0Ed}`] First order moment; see also 5.8.8.2 (2) [:math:`kNm`].
        beta : float
            [:math:`β`] Factor which depends on distribution of 1st and 2nd order moments, see 5.8.7.3 (2)-(3) [-].
        n_ed : KN
            [:math:`N_{Ed}`] Design value of axial load [:math:`kN`].
        n_b : KN
            [:math:`N_{B}`] Buckling load based on nominal stiffness [:math:`kN`].
        """
        super().__init__()
        self.m_0ed = m_0ed
        self.beta = beta
        self.n_ed = n_ed
        self.n_b = n_b

    @staticmethod
    def _evaluate(
        m_0ed: KNM,
        beta: DIMENSIONLESS,
        n_ed: KN,
        n_b: KN,
    ) -> KN:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            m_0ed=m_0ed,
            beta=beta,
            n_b=n_b,
        )
        raise_if_less_or_equal_to_zero(n_ed=n_ed)

        # When n_b / n_ed is equal to 1, the formula will divide by zero.
        # The spirit of the equation is to raise the second order moment to infinity when n_b / n_ed is equal to 1.
        # This is not possible in numeric calculations, so we will use a large number instead.
        return 1e9 if n_b == n_ed else m_0ed * (1 + beta / (n_b / n_ed - 1))

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.28."""
        return LatexFormula(
            return_symbol=r"M_{Ed}",
            result=f"{self:.3f}",
            equation=r"M_{0Ed} \cdot \left(1 + \frac{\beta}{\frac{N_{B}}{N_{Ed}} - 1}\right)",
            numeric_equation=rf"{self.m_0ed:.3f} \cdot \left(1 + \frac{{{self.beta:.3f}}}{{\frac{{{self.n_b:.3f}}}{{{self.n_ed:.3f}}} - 1}}\right)",
            comparison_operator_label="=",
            unit="kNm",
        )
